import math
import tkinter
from tkinter import ttk, filedialog, colorchooser, messagebox

import sv_ttk


###Time Me
import time

def time_me(f):
    '''Decorator function to time functions' runtime in ms'''
    def wrapper(*args, **kwargs):
        start = time.time()
        res = f(*args, **kwargs)
        print(f'function: {f.__name__} took {(time.time()-start)*1000:.4f}ms')
        return res
    return wrapper


###extract data
import re

import numpy as np

def extract_data(file):
    """
    @brief: This function takes in a .obj file and returns a list containing
            the indexes of the vertexes that make up the face, and a dict
            that contains the index of the vertex, along with its coordinates

    @param: .obj file

    @ret  : verticies (the dict containing coordinates)
    @ret  : faces (the list containing the vertexes' indexes)
    """

    verticies = np.empty((1, 3))
    faces = []

    # Read more about how waveform (.obj) files are structured to understand
    # how this code exactly works, but shortly:
    #   * If the line starts with a "v", then that's a vertex and what follows is
    #     its X, Y, Z coordinates (foats)
    #   * If the line starts with a "f", then that's a face and what follows is
    #     the list of verticies to be connected to create a face
    #     (formatted a bit strangely though, I recommend checking an example)

    for line in file.readlines():
        if line[0:2] == "v ":
            pt = np.array([[float(x) for x in re.findall("[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", line)]])
            verticies = np.concatenate([verticies, pt])
        elif line[0:2] == "f ":
            faces.append([int(vertex.split("/")[0]) for vertex in line[2:-2].split(' ')])
    return verticies, faces

###MAIN

class Canvas3D(tkinter.Canvas):
    def __init__(self, frame, parent=None):
        if parent is None:
            parent=tkinter.Tk()
        self.parent=parent
        #inherit
        try:
            self.assets=self.parent.assets
        except:
            self.assets=Assets()
        try:
            self.data=self.parent.data
        except:
            self.data=Shared_Data()
        self.parent_frame=frame
        self._geometry_handler = Geometry(self.CANVAS_WIDTH, self.CANVAS_HEIGHT)
        super().__init__(self.parent_frame)
        self.pack(expand=1, fill='both')
        self._canvas.bind("<Configure>", self.__resized)
        self._changed = False

        self._geometry_handler.reset_rotation()  
        self._file_name.set(file_path.split('/')[-1])
        with open(file_path) as file:
            self._geometry_handler.upload_object(*obj_files_handler.extract_data(file))
            self._file_exists = True


    def __get_canvas_shape(self):
        """returns the shape of the canvas holding the visualized frame"""
        self.update()
        return self._canvas.winfo_width(), self._canvas.winfo_height()

    def __resized(self, *args):
        '''Callback to the window resize events'''
        w, h = self.__get_canvas_shape()
        if self._canvas_w != w or self._canvas_h != h:
            # Keep the object in the middle of the canvas
            self._geometry_handler.update_position((w-self._canvas_w)//2, (h-self._canvas_h)//2)
            self._canvas_w = w
            self._canvas_h = h
            self.__changed()

    def __changed(self, *args):
        '''Signal to the rendering function that something has changed in the object'''
        self._changed = True


    def render(self):

        self.__set_rotations()
        # self.__set_zoom()

        #Delete all the previous points and lines in order to draw new ones
        self._canvas.delete("all")
        self.__update_colors()
        self.__draw_object()
        self._changed = False

    def __set_zoom(self):
        self._geometry_handler.set_zoom(self._zoom_slider.get())

    def __set_rotations(self):
        '''Set the required rotations for the geometry handler'''
        x, y, z = self._geometry_handler.orientation
        self._geometry_handler.reset_rotation(
            x=x, 
            y= y, 
            z=z
        )

    def __draw_point(self, point: 'tuple(int, int)') -> None:
        '''Draw a point on the canvas'''
        self._canvas.create_oval(point[0], point[1],
                                    point[0], point[1],
                                    width=self.POINT_SIZE,
                                    fill=self.POINT_COLOR)
    
    @time_me
    def __draw_faces(self, points: dict) -> None:
        ''''''
        for face in self._geometry_handler.faces:
            # Grab the points that make up that specific face
            to_draw = [points[f] for f in face]
            
            #for point in to_draw:
            #    if(point[0] < 0 or
            #       point[1] < 0 or
            #       point[0] > self.CANVAS_WIDTH or
            #       point[1] > self.CANVAS_HEIGHT
            #    ):
            #        continue # Don't draw points that are out of the screen
            #    # This is the slowest part of the GUI
            #    self.__draw_point(point)

            self._canvas.create_polygon(to_draw, outline=self._line_color_holder, fill=self._fill_color_holder)

    def __draw_object(self):
        '''Draw the object on the canvas'''
        projected_points = self._geometry_handler.transform_object()
        self.__draw_faces(projected_points)


####GEOMETRY FILE
import math
import numpy as np
import numba

@numba.njit(nogil=True, cache=True, fastmath=True)
def matmul(A: np.ndarray, B: np.ndarray) -> np.ndarray:
    '''
    @brief: Matrix multiplication from scratch
    @param A: First matrix 
    @param B: Second matrix
    @Note: A, and B must be multiplicable matrices; that must have a common dimension
    @return C: Product of A and B
    '''
    rows, cols = A.shape[0], B.shape[1]
    C = np.zeros((rows, cols))
    for i in range(rows):
        for j in range(cols):
            for k in range(rows):
                C[i, j] += A[i, k] * B[k, j]
    return C

@numba.njit(nogil=True, cache=True)
def max_3d_array(arr: np.ndarray, axis: int) -> float:
    '''
    Find the maximum in a miltidimensional array, in the provided axis
    '''
    max_ = -np.inf
    for i in arr:
        if i[axis] >= max_: 
            max_ = i[axis] 
    return max_

@numba.njit(nogil=True, cache=True)
def min_3d_array(arr: np.ndarray, axis: int) -> float:
    '''
    Find the minimum in a miltidimensional array, in the provided axis
    '''
    min_ = np.inf
    for i in arr:
        if i[axis] <= min_: 
            min_ = i[axis] 
    return min_

class Geometry:
    '''
    Geometry handling class (linear algebra)
    '''
    OBJECT_SCALE = 2000 # Maybe make this dynamic depending on the object size
    
    def __init__(self, canvas_width: int, canvas_height: int) -> None:
        '''
        '''
        self._obj_position = np.array((canvas_width//2, canvas_height//2))
        self._zoom = 50.0
        self._angle_x = 0.0
        self._angle_y = 0.0
        self._angle_z = 0.0
        self._faces = None
        self._verticies = None

    def upload_object(self, verts: np.ndarray, faces: list) -> None:
        '''Uploads the verticies and faces to manipulate'''
        self._verticies = self.__normalize_3d_array(verts, axis=0)
        self._faces = faces

    def update_position(self, x: int, y: int) -> None:
        '''Update x, y position of the object'''
        self._obj_position[0] += x
        self._obj_position[1] += y
    
    @time_me
    def transform_object(self) -> 'list(list(int, int))':
        '''Retur the points of the object transformed according to the current pose'''
        rot_x, rot_y, rot_z = self.__calculate_rot_matrix()
        projected_points = []
        for pt in self._verticies:
            x, y = self.__transform_point(pt, rot_x, rot_y, rot_z, self._zoom, self._obj_position, self.OBJECT_SCALE)
            projected_points.append([x, y])
        return projected_points

    @property
    def faces(self) -> list:
        '''Get the faces formed between the points'''
        return self._faces

    @property
    def zoom(self) -> int:
        '''Get the current zoom value'''
        return self._zoom

    @property
    def orientation(self) -> 'tuple(float, float, float)':
        '''Returns the object's current angles'''
        return self._angle_x, self._angle_y, self._angle_z

    def set_zoom(self, zoom: float) -> None:
        '''Set the new zoom value'''
        self._zoom = zoom

    def step_rotation(self, 
                      x: float = 0.0,
                      y: float = 0.0,
                      z: float = 0.0
    ) -> None:
        '''Increment the orientation of the object on its axis'''
        self._angle_x += x
        self._angle_y += y
        self._angle_z += z

    def reset_rotation(self, 
                       x: float = None,
                       y: float = None,
                       z: float = None
    ) -> None:
        '''Reset the rotation to a specific position, if provided, else to 0'''
        self._angle_x = 0 if x is None else x
        self._angle_y = 0 if y is None else y
        self._angle_z = 0 if z is None else z

    @staticmethod
    @numba.njit(nogil=True, cache=True, fastmath=True)
    def __normalize_3d_array(arr: np.ndarray, 
                             range: 'tuple(float, float)' = (-1, 1),
                             axis: int = 2
    ) -> np.ndarray:
        '''
        @brief: Normalize an array values within a range based on a specified axis
        @param arr: The array to be normalized
        @param range: Normalized values range (min, max)
        @param axis: the axis to normalize based on
        @return arr: The normalized array
        '''
        mnx = min_3d_array(arr, 0)
        mxx = max_3d_array(arr, 0)
        mnz = min_3d_array(arr, 2)
        mxz = max_3d_array(arr, 2)
        mny = min_3d_array(arr, 1)
        mxy = max_3d_array(arr, 1)

        if axis == 0:
            diff = mxx - mnx
        elif axis == 1:
            diff = mxy - mny
        else:
            diff = mxz - mnz
        
        for pt in arr:
            pt[0] = (((pt[0]-mnx)*(range[1]-range[0]))/diff) + range[0]
            pt[1] = (((pt[1]-mny)*(range[1]-range[0]))/diff) + range[0]
            pt[2] = (((pt[2]-mnz)*(range[1]-range[0]))/diff) + range[0]    
        return arr

    @staticmethod
    @numba.njit(nogil=True, cache=True, fastmath=True)
    def __transform_point(point: np.ndarray, 
                          rotation_x: np.ndarray, 
                          rotation_y: np.ndarray, 
                          rotation_z: np.ndarray,
                          zoom: float,
                          obj_position: 'list[int, int]',
                          obj_scale: int
    ) -> 'tuple(int, int)':
        '''
        @brief: Rotate the point in 3axis according to the provided rotation matrices
        @param point: 3D point
        @param rotation_x: Rotation matrix on X axis
        @param rotation_y: Rotation matrix on Y axis
        @param rotation_z: Rotation matrix on Z axis
        @param zoom: Zoom value
        @param obj_position: Object position within the screen
        @param obj_scale: Object scale
        @return transformed point: 2D tranformed projection of the 3D point
        '''
        # Rotate point on the Y, X, and Z axis respectively
        rotated_2d = matmul(rotation_y, point.reshape((3, 1)))
        rotated_2d = matmul(rotation_x, rotated_2d)
        rotated_2d = matmul(rotation_z, rotated_2d)

        # Project 3D point on 2D plane
        z = 0.5 / (zoom - rotated_2d[2][0])
        projection_matrix = np.array(((z, 0, 0), (0, z, 0)))
        projected_2d = matmul(projection_matrix, rotated_2d)

        x = int(projected_2d[0][0]*obj_scale) + obj_position[0]
        # The (-) sign in the Y is because the canvas' Y axis starts from Top to Bottom, 
        # so without the (-) sign, our objects would be presented upside down
        y = -int(projected_2d[1][0]*obj_scale) + obj_position[1]

        return x, y

    def __calculate_rot_matrix(self) -> 'tuple(np.array, np.array, np.array)':
        '''
        Calculate the rotation matrices on X, Y, and Z axis 
        that correspond to the current requested rotation
        '''
        rotation_x = np.array(
            (
                (1,               0        ,               0         ),
                (0, math.cos(self._angle_x), -math.sin(self._angle_x)),
                (0, math.sin(self._angle_x),  math.cos(self._angle_x))
            )
        )

        rotation_y = np.array(
            (
                (math.cos(self._angle_y), 0, -math.sin(self._angle_y)),
                (            0          , 1,             0           ),
                (math.sin(self._angle_y), 0,  math.cos(self._angle_y))
            )
        )

        rotation_z = np.array(
            (
                (math.cos(self._angle_z), -math.sin(self._angle_z), 0),
                (math.sin(self._angle_z),  math.cos(self._angle_z), 0),
                 (           0           ,              0          , 1)
            )
        )
        return rotation_x, rotation_y, rotation_z




import tkinter
import tkinter.font
import tkinter.ttk
from tkcalendar import Calendar


class highlable_calendar(Calendar):
    def __init__(self, *args, parent=None, **kwargs):
        self.data_available_list=[]
        self.parent=parent
        super(highlable_calendar, self).__init__(*args, **kwargs)
        self.style.configure('data.%s.TLabel' % self._style_prefixe, background="OliveDrab1",
                             foreground=self._properties.get('normalforeground'))
        self.bind("<<CalendarSelected>>", self.day_selection_callback)

    def _display_calendar(self):
        super()._display_calendar()
        for item in self.data_available_list:
            self.highlight(item)

    def _remove_selection(self):
        if self._sel_date is not None:
            super()._remove_selection()
            if self._sel_date in self.data_available_list:
                self.highlight(self._sel_date)

    def highlight(self,date):
        w, d = self._get_day_coords(date)
        if w is not None:
            label = self._calendar[w][d]
            if label.cget('text'):
                label.configure(style='data.%s.TLabel' % self._style_prefixe)

    def add_to_data(self, date):
        self.configure(state="normal")
        self.data_available_list.append(date)
        if date != self._sel_date:
            self.highlight(date)


    def unhighlight(self):
        self.configure(state="disabled")
        self._display_calendar()

    def day_selection_callback(self, event=None):
        if self.parent is not None:
            if self._sel_date in self.data_available_list:
                self.parent.event_generate("<<UpdatePlotData>>")


import tkinter
import tkinter.font
import tkinter.ttk
import os, sys, webbrowser, platform
import time
import traceback
import re


class SHARED():
    def __init__(self):
        self.Turbidity=tkinter.DoubleVar()
        self.PH=tkinter.DoubleVar()
        self.Battery=tkinter.DoubleVar()
        self.Temperature=tkinter.DoubleVar()
        self.Conductivity=tkinter.DoubleVar()
        self.Sonar=tkinter.DoubleVar()
        self.Date=tkinter.StringVar()
        self.Date.set("                  ")



from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, ConstantKernel, Matern, WhiteKernel

import torch
import gpytorch

from botorch.models.gp_regression import FixedNoiseGP 
from botorch.fit import fit_gpytorch_mll

import numpy as np

class GaussianProcessScikit:

	def __init__(self, scenario_map, initial_lengthscale=5.0, kernel_bounds=(0.1, 100)):
		""" Gaussian Process with Scikit-Learn to predict the map """

		# Assign input variables #
		self.scenario_map = scenario_map

		# Set the initial lengthscale and kernel bounds
		self.initial_lengthscale = initial_lengthscale
		self.kernel_bounds = kernel_bounds

		# Calculate visitable cells #
		self.visitable_indexes = np.where(self.scenario_map == 1) # visitable cells indexes
		self.X_cells = np.vstack(self.visitable_indexes).T # visitable cells coords

		self.reset()

	def fit_gp(self, X_new, y_new, variances_new):
		""" Fit Gaussian Process to data """

		# Add new measurements or update existing if new variance better than older #
		self.all_train_measures_dict.update({tuple(X): (y, variance) for X, y, variance in zip(X_new, y_new, variances_new) if tuple(X) not in self.all_train_measures_dict or variance < self.all_train_measures_dict[tuple(X)][1]})

		# Get data from dictionary:  X_meas: stored in keys, y_meas: stored in first column of values tuple, variance_meas: stored in second column of values tuple #
		X_meas, y_meas, variance_meas = zip(*[(key, value[0], value[1]) for key, value in self.all_train_measures_dict.items()])

		# Update internal gp alpha variable (variance) #
		self.gp.alpha = np.array(variance_meas)

		# FIT #
		self.gp.fit( np.array(X_meas), np.array(y_meas)) 

	def predict_gt(self):
		""" Ground truth prediction for visitable cells (X_cells) """
		
		model_out, uncertainty_out = self.gp.predict(self.X_cells, return_std=True)

		# Assign GP predictions to corresponding map cells, "flatten" to match dimensions
		self.model_map[self.visitable_indexes] = model_out.flatten()
		self.uncertainty_map[self.visitable_indexes] = uncertainty_out.flatten()

		return self.model_map, self.uncertainty_map

	def reset(self):

		# Initialize dataset dictionary: keys = tuple(position of measure), values = tuple(measure, variance) #
		self.all_train_measures_dict = {}

		# Define empty maps to predict the model #
		self.model_map = np.zeros_like(self.scenario_map)
		self.uncertainty_map = np.zeros_like(self.scenario_map)

		# Define Gaussian Process with Kernel #
		# self.kernel = ConstantKernel(1) #* Matern(length_scale=5, length_scale_bounds=(1, 1000)) + WhiteKernel(0.005, noise_level_bounds=(0.00001, 0.1))
		self.kernel = RBF(length_scale=self.initial_lengthscale, length_scale_bounds=self.kernel_bounds) # RBF kernel
		self.gp = GaussianProcessRegressor(kernel=self.kernel, alpha=0.001, n_restarts_optimizer=5)


class GaussianProcessGPyTorch:
	""" Gaussian Process with GPyTorch to predict the map """

	def __init__(self, scenario_map, initial_lengthscale=5.0, kernel_bounds=(0.1, 100), training_iterations=50, scale_kernel:bool=True, device='cpu', mean=0):

		# Set input map #
		self.scenario_map = scenario_map
		self.mean=0
		# Set the initial lengthscale and kernel bounds
		self.initial_lengthscale = initial_lengthscale
		self.kernel_bounds = kernel_bounds
		
		# Set the training iterations
		self.training_iterations = training_iterations

		# Set other config #
		self.scale_kernel = scale_kernel
		self.device = device

		# Calculate visitable cells #
		self.visitable_indexes = np.where(self.scenario_map == 1) # visitable cells indexes
		self.X_cells = np.vstack(self.visitable_indexes).T # visitable cells coords

		# Call reset function to create regression model #
		self.reset()

	def fit_gp(self, X_new, y_new, variances_new, verbose=False):#, train_x:np.ndarray, train_y:np.ndarray, verbose=False):

		# Add new measurements or update existing if new variance better than older #
		self.all_train_measures_dict.update({tuple(X): (y, variance) for X, y, variance in zip(X_new, y_new, variances_new) if tuple(X) not in self.all_train_measures_dict or variance < self.all_train_measures_dict[tuple(X)][1]})

		# Get data from dictionary:  X_meas: stored in keys, y_meas: stored in first column of values tuple, variance_meas: stored in second column of values tuple #
		X_meas, y_meas, variance_meas = zip(*[(key, value[0], value[1]) for key, value in self.all_train_measures_dict.items()])

		# Convert train data and noise to TENSOR #
		self.train_x = torch.FloatTensor(X_meas).to(self.device)
		self.train_y = torch.FloatTensor(y_meas).to(self.device)
		self.train_noise = torch.FloatTensor(variance_meas).to(self.device)

		# Update with new train data and noise (samples variance) # 
		self.model.set_train_data(self.train_x, self.train_y, strict=False)
		self.likelihood.noise = self.train_noise

		# Find optimal model hyperparameters (training) #
		self.model.train()
		self.likelihood.train()

		# Iterate over training iterations
		converged = 0
		it = 0
		#kernel_lengthscale = self.model.covar_module.base_kernel.lengthscale.item()
		loss_ant = None
		while it < self.training_iterations and not converged:
			# Zero backprop gradients
			self.optimizer.zero_grad()
			# Get output from model
			output = self.model(self.train_x)
			# Calc loss and backprop gradients
			loss = -self.mll(output, self.train_y)
			loss.backward()
			if verbose:
				print('Iter %d/%d - Loss: %.3f   lengthscale: %.3f   noise: %.3f' % (
					it + 1, self.training_iterations, loss.item(),
					self.model.covar_module.base_kernel.lengthscale.item(),
					self.model.likelihood.noise.item()
				))

			self.optimizer.step()

			if loss_ant is None:
				loss_ant = loss.item()
			else:
				converged = np.abs(loss.item() - loss_ant) < 0.01

			#kernel_lengthscale = self.model.covar_module.base_kernel.lengthscale.item()

			it += 1

	def predict_gt(self): #, eval_x:np.ndarray):
		""" Evaluate the model """

		eval_x = torch.FloatTensor(self.X_cells).to(self.device)

		# Set into eval mode
		self.model.eval()
		self.likelihood.eval()

		# Make predictions
		with torch.no_grad(), gpytorch.settings.fast_pred_var():
			# observed_pred = self.likelihood(self.model(eval_x))
			observed_pred = self.model(eval_x)

		# Extract predictions #
		model_out = observed_pred.mean.detach().cpu().numpy()
		uncertainty_out = observed_pred.stddev.detach().cpu().numpy()

		# Assign GP predictions to corresponding map cells, "flatten" to match dimensions
		self.model_map[self.visitable_indexes] = model_out.flatten()
		self.uncertainty_map[self.visitable_indexes] = uncertainty_out.flatten()
		
		return self.model_map, self.uncertainty_map

	def reset(self):

		# Initialize dataset dictionary: keys = tuple(position of measure), values = tuple(measure, variance) #
		self.all_train_measures_dict = {}

		# Define empty maps to predict the model #
		self.model_map = np.zeros_like(self.scenario_map)
		self.uncertainty_map = np.zeros_like(self.scenario_map)

		# Initialize train and noise tensors #
		self.train_x = torch.FloatTensor([]).to(self.device)
		self.train_y = torch.FloatTensor([]).to(self.device)
		self.train_noise = torch.FloatTensor([]).to(self.device)

		# Set likelihood and model #
		# self.likelihood = gpytorch.likelihoods.GaussianLikelihood().to(self.device)
		self.likelihood = gpytorch.likelihoods.FixedNoiseGaussianLikelihood(noise=self.train_noise).to(self.device) # FixedNoiseGaussianLikelihood to add measurements noise
		self.model = ExactGPModel(train_x=self.train_x, train_y=self.train_y, likelihood=self.likelihood, initial_lengthscale=self.initial_lengthscale, kernel_bounds=self.kernel_bounds, scale_kernel=self.scale_kernel, mean=self.mean).to(self.device)

		# Set "loss" for GPs - the marginal log likelihood (mll)
		self.mll = gpytorch.mlls.ExactMarginalLogLikelihood(self.likelihood, self.model)

		# Set up to use the adam optimizer
		self.optimizer = torch.optim.Adam(self.model.parameters(), lr=0.1)

""" ExactGPModel is a class that inherits from gpytorch.models.ExactGP. It is used to create a GP model for regression."""
class ExactGPModel(gpytorch.models.ExactGP):

	def __init__(self, train_x, train_y, likelihood, initial_lengthscale=5.0, kernel_bounds=(0.5, 10), scale_kernel:bool=True, mean = 0):
		super().__init__(train_x, train_y, likelihood)

		# Declare the mean and covariance modules #
		self.mean_module = gpytorch.means.ConstantMean()
		self.mean_module.initialize(constant=mean)
		# self.mean_module = gpytorch.means.ZeroMean()
		
		# Declare the covariance module and set the initial lengthscale and constraints #
		#First, check if the initial lengthscale is in the bounds
		assert kernel_bounds[0] <= initial_lengthscale <= kernel_bounds[1], "The initial lengthscale is not in the bounds"
		
		if scale_kernel == True:
			# Declare the kernel #
			self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel(lengthscale_constraint=gpytorch.constraints.Interval(*kernel_bounds)))
			# Set the initial lengthscale #
			self.covar_module.base_kernel.lengthscale = initial_lengthscale
		else:
			# Declare the kernel #
			self.covar_module = gpytorch.kernels.RBFKernel(lengthscale_constraint=gpytorch.constraints.Interval(*kernel_bounds))
			# Set the initial lengthscale #
			self.covar_module.lengthscale = initial_lengthscale

	def forward(self, x):
		mean_x = self.mean_module(x)
		covar_x = self.covar_module(x)
		return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class GaussianProcessBoTorch:
	""" Gaussian Process with BoTorch to predict the map """

	def __init__(self, scenario_map, initial_lengthscale=5.0, kernel_bounds=(0.1, 100), device='cpu'):

		# Set input map #
		self.scenario_map = scenario_map

		# Set the initial lengthscale and kernel bounds
		self.initial_lengthscale = initial_lengthscale
		self.kernel_bounds = kernel_bounds

		# Set calculate device #
		self.device = device

		# Calculate visitable cells #
		self.visitable_indexes = np.where(self.scenario_map == 1) # visitable cells indexes
		self.X_cells = np.vstack(self.visitable_indexes).T # visitable cells coords

		# Call reset function to create regression model #
		self.reset()

	def fit_gp(self, X_new, y_new, variances_new, verbose=False):#, train_x:np.ndarray, train_y:np.ndarray, verbose=False):

		# Add new measurements or update existing if new variance better than older #
		self.all_train_measures_dict.update({tuple(X): (y, variance) for X, y, variance in zip(X_new, y_new, variances_new) if tuple(X) not in self.all_train_measures_dict or variance < self.all_train_measures_dict[tuple(X)][1]})

		# Get data from dictionary:  X_meas: stored in keys, y_meas: stored in first column of values tuple, variance_meas: stored in second column of values tuple #
		X_meas, y_meas, variance_meas = zip(*[(key, value[0], value[1]) for key, value in self.all_train_measures_dict.items()])

		# Convert train data and noise to TENSOR #
		self.train_x = torch.FloatTensor(X_meas).to(self.device)
		self.train_y = torch.FloatTensor(y_meas).to(self.device)
		self.train_noise = torch.FloatTensor(variance_meas).to(self.device)

		# Update with new train data and noise (samples variance) # 
		self.model.set_train_data(self.train_x, self.train_y, strict=False)
		self.model.likelihood.noise = self.train_noise
		self.model.likelihood._aug_batch_shape = len(self.train_noise)

		fit_gpytorch_mll(self.mll)

	def predict_gt(self): #, eval_x:np.ndarray):
		""" Evaluate the model """

		eval_x = torch.FloatTensor(self.X_cells).to(self.device)

		# Make predictions
		with torch.no_grad():
			posterior = self.model.posterior(eval_x)
			model_out = posterior.mean.cpu()
			uncertainty_out = posterior.stddev.cpu()
    
		# Assign GP predictions to corresponding map cells, "flatten" to match dimensions
		self.model_map[self.visitable_indexes] = model_out.flatten()
		self.uncertainty_map[self.visitable_indexes] = uncertainty_out.flatten()
		
		return self.model_map, self.uncertainty_map

	def reset(self):

		# Initialize dataset dictionary: keys = tuple(position of measure), values = tuple(measure, variance) #
		self.all_train_measures_dict = {}

		# Define empty maps to predict the model #
		self.model_map = np.zeros_like(self.scenario_map)
		self.uncertainty_map = np.zeros_like(self.scenario_map)

		# Initialize train and noise tensors #
		self.train_x = torch.empty((0, 2), device=self.device)
		self.train_y = torch.empty((0,1), device=self.device)
		self.train_noise = torch.empty((0,1), device=self.device)

		# Set likelihood and model #
		self.model = FixedNoiseGP(train_X=self.train_x, train_Y=self.train_y, train_Yvar=self.train_noise).to(self.device)

		# Set "loss" for GPs - the marginal log likelihood (mll)
		self.mll = gpytorch.mlls.ExactMarginalLogLikelihood(self.model.likelihood, self.model)

		# Set up to use the adam optimizer
		self.optimizer = torch.optim.Adam(self.model.parameters(), lr=0.1)



import tkinter
import tkinter.font
import tkinter.ttk
import os, sys
from tkintermapview import TkinterMapView
from datetime import datetime, date
import pandas as pd
import tkinter
sys.path.append('../')
from ..submodules.database import Database
from ..shared import *
from ..assets import *
from ..widgets.highlable_calendar import *
from ..widgets.labels import *
from ..widgets.map_markers import *
from ..widgets.checkbox import *
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg,
                                               NavigationToolbar2Tk)


class TRASHTAB(tkinter.PanedWindow):
    def __init__(self, parent=None):
        if parent is None:
            parent=tkinter.Tk()
        self.parent=parent
        self.trash_markers=[]
        #inherit
        try:
            self.assets=self.parent.assets
            self.shared=self.parent.shared
        except:
            self.assets=Assets()
            self.shared=SHARED()

        super().__init__(orient="horizontal")

        self.database= Database()
        self.date=date.today()
        self.download_loop_object = None
        self.refreshed = False
        self.busy=False
        self.sm_state=0
        self.path_map=None
        self.path_db = pd.DataFrame()
        self.trash_db = pd.DataFrame()
        self.wqp_db = pd.DataFrame()


        self.trash_tab()

    def trash_tab(self):
        self.GPS_panel = tkinter.PanedWindow(orient="vertical")
        self.add(self.GPS_panel)
        
        #create map
        self.map_widget = TkinterMapView(corner_radius=0, height=int(self.parent.screenheight*0.72)+1,database_path=self.assets.map_tiles) #this widget has no automatic size
        self.map_widget.set_tile_server("https://mt0.google.com/vt/lyrs=s&hl=en&x={x}&y={y}&z={z}&s=Ga", max_zoom=22)
        self.GPS_panel.add(self.map_widget)
        self.topleft=[37.420088,-6.001346]
        self.bottomright=[37.418399,-5.997504]
        self.map_widget.fit_bounding_box(self.topleft, self.bottomright)


        self.trashes=[]
        self.ships=[]

        ##create info bar
        gps_data= tkinter.PanedWindow(orient="horizontal",height=1)
        self.GPS_panel.add(gps_data, minsize=60)
        #scale
        self.timeline = tkinter.Scale(gps_data, from_=0, to=0, orient="horizontal",command=self.go_to_time, bg="white")
        gps_data.add(self.timeline,width=5000)
        self.play_callback=None
        #play and center buttons
        aux=tkinter.Frame(gps_data, height=1, borderwidth=1, bg="white") 
        buttonsframe=tkinter.Frame(aux, height=1, borderwidth=1) #for date and play/center
        buttonsframe.pack(side="top")
        self.playbuttom = tkinter.Button(buttonsframe, image=self.assets.icon_play, command=self.play, width=35, height=35, bg="white")
        self.playbuttom.pack(side="left")
        self.play_status=False
        self.playbuttom["state"] = "disabled"
        self.center_buttom = tkinter.Button(buttonsframe, text="Center", command=self.center_map, width=10, height=1, bg="white")
        self.center_buttom.pack(side="right",expand="true",fill="both")
        self.speed_buttom = tkinter.Label(buttonsframe, text="x1", relief="raised" , width=5, height=1, bg="white")
        self.speed_buttom.bind("<Button-1>", self.increase_speed)
        self.speed_buttom.bind("<Button-3>", self.decrease_speed)
        self.speed=1
        self.speed_buttom.pack(side="right",expand="true",fill="y")


        #colorbar scale plot
        self.colorbar_fig, self.colorbar_ax = plt.subplots(figsize=(1,20), layout='constrained')
        self.colorbar = FigureCanvasTkAgg(self.colorbar_fig, master=self)
        self.add(self.colorbar.get_tk_widget(), minsize=30)
        
        #date
        self.date_var=tkinter.StringVar()
        self.date_label=tkinter.Label(aux, textvariable=self.date_var, width=int(self.parent.screenwidth*0.08), height=1, font=tkinter.font.Font(weight='bold', size=12))
        self.date_label.pack(side="bottom", padx=0, pady=2,anchor="c")
        self.mouse_position = tkinter.StringVar()
        self.mouse_position.set("Lat: --, Lon: --")
        self.mouse_latlon=tkinter.Label(aux , textvariable=self.mouse_position, width=int(self.parent.screenwidth*0.12), height=1, font=tkinter.font.Font(weight='bold', size=12))
        self.mouse_latlon.pack(side="top", padx=0, pady=2,anchor="c", fill="both")
        self.map_widget.canvas.bind("<Motion>", self.update_mouse_position)
        self.map_widget.canvas.bind("<Leave>", self.clear_mouse_position)
        gps_data.add(aux,minsize=290)
        
        #map selection
        self.map_selection=tkinter.Frame(self)
        self.map_selection.pack_propagate(False)
        self.add(self.map_selection, minsize=240)

        #calendar
        self.playback_selector_frame=tkinter.Frame(self.map_selection)
        self.cal = highlable_calendar(self.playback_selector_frame, selectmode = 'day', parent=self)
        self.cal.selection_set(self.date)
        self.bind("<<UpdatePlotData>>", self.download_database)

        self.cal.pack(expand="false",fill="none")
        self.refresh_buttom = tkinter.Button(self.playback_selector_frame, text="refresh", command=self.refresh_database, width=8, height=1,font=tkinter.font.Font(weight='bold', size=18))
        self.refresh_buttom.pack(side="top",expand="false",fill="x")


        self.playback_selector_frame.pack()

        self.checkboxes=[]
        for i in range(len(self.database.sensors)): #for as many sensors
            aux= Checkbox(self.map_selection, text=self.database.sensors[i], width=20, command=lambda a=i: self.draw_map(a))
            aux.pack(side="top")
            self.checkboxes.append(aux)
            aux["state"] = "disabled"
        
        self.data_points_label=tkinter.StringVar()
        a= tkinter.Frame(self.map_selection, padx=10, pady=2)
        a.pack(side='top', fill='both', expand=True,)
        a.pack_propagate(False)
        tkinter.Label(a, text="Mission Length: ", font=tkinter.font.Font(weight='bold')).pack(side="left", fill='both', padx=0, pady=2)
        tkinter.Label(a, textvariable=self.data_points_label).pack(side="left", fill='both', padx=0, pady=2)

        self.wqp_points_label=tkinter.StringVar()
        a= tkinter.Frame(self.map_selection, padx=10, pady=2)
        a.pack(side='top', fill='both', expand=True,)
        a.pack_propagate(False)
        tkinter.Label(a, text="wqp points: ", font=tkinter.font.Font(weight='bold')).pack(side="left", fill='both', padx=0, pady=2)
        tkinter.Label(a, textvariable=self.wqp_points_label).pack(side="left", fill='both', padx=0, pady=2)


        self.trash_points_label=tkinter.StringVar()
        a= tkinter.Frame(self.map_selection, padx=10, pady=2)
        a.pack(side='top', fill='both', expand=True,)
        a.pack_propagate(False)
        tkinter.Label(a, text="Trash Pressence: ", font=tkinter.font.Font(weight='bold')).pack(side="left", fill='both', padx=0, pady=2)
        tkinter.Label(a, textvariable=self.trash_points_label).pack(side="left", fill='both', padx=0, pady=2)

        self.status_label=tkinter.StringVar()
        a= tkinter.Frame(self.map_selection, padx=10, pady=2)
        a.pack(side='top', fill='both', expand=True,)
        a.pack_propagate(False)
        tkinter.Label(a, text="Status", font=tkinter.font.Font(weight='bold')).pack(side="left", fill='both', padx=0, pady=2)
        tkinter.Label(a, textvariable=self.status_label).pack(side="left", fill='both', padx=0, pady=2)
        self.status_label.set("Off")

    def update_mouse_position(self, event=None):
        try:
            latitude, longitude = self.map_widget.convert_canvas_coords_to_decimal_coords(event.x, event.y)
            self.mouse_position.set(f"Lat: {latitude:.6f}, Lon: {longitude:.6f}")
        except Exception:
            pass

    def clear_mouse_position(self, event=None):
        self.mouse_position.set("Lat: --, Lon: --")


    def play(self):
        if self.play_status:
            self.playbuttom.config(image=self.assets.icon_play)
            self.parent.after_cancel(self.play_callback)
            self.play_status=False
            self.status_label.set("Paused")
        else:
            if self.timeline.get() == len(self.path_db):
                self.timeline.set(0)
            self.playbuttom.config(image=self.assets.icon_pause)
            self.play_callback=self.parent.after(200//self.speed,self.execute_time)
            self.play_status=True
            self.status_label.set("Playing")



    def go_to_time(self, value):
        value=int(value)
        if value>=len(self.path_db):
            return
        if value == 0: #reset
            self.map_widget.delete_all_marker()
            self.map_widget.delete_all_path()
            self.ships=[]
            self.trashes=[]
            if len(self.wqp_db) >0:
                self.path_map.show_up_to(0)

        #get next time
        timestep=self.path_db.index.values[value]
        self.date_var.set(pd.to_datetime(timestep).strftime("%Y-%m-%d %H:%M:%S"))
        if len(self.wqp_db) >0:
            self.path_map.show_up_to(timestep)
        if len(self.ships)==0:
            self.ships =[SHIP("ASV1", self.path_db.at[timestep,"Latitude"], self.path_db.at[timestep,"Longitude"], rotation=-self.path_db.at[timestep,"Heading"], parent=self)]
        else:
            self.ships[0].update_rotation(-self.path_db.at[timestep,"Heading"])
            self.ships[0].set_position(self.path_db.at[timestep,"Latitude"], self.path_db.at[timestep,"Longitude"])
        if timestep in self.trash_db.index.values:
            if len(self.trashes)==0:
                self.trashes=[TRASH("Trash_1", self.trash_db.at[timestep,"Latitude_Obj"],self.trash_db.at[timestep,"Longitude_Obj"], parent=self)]
            else:
                self.trashes[0].update_position(self.trash_db.at[timestep,"Latitude_Obj"],self.trash_db.at[timestep,"Longitude_Obj"])
        else:
            if len(self.trashes)!=0:
                self.trashes[0].hide_image(True)

            
    def execute_time(self):
        if self.timeline.get() == len(self.path_db) and self.play_status:
            self.play() #stop time
            return
        self.timeline.set(self.timeline.get()+1)
        self.play_callback=self.parent.after(200//self.speed,self.execute_time)
        
    def increase_speed(self, event=None):
        if self.speed<8:
            self.speed+=1
        self.speed_buttom.config(text=f"x{self.speed}")

    def decrease_speed(self, event=None):
        if self.speed>1:
            self.speed-=1
        self.speed_buttom.config(text=f"x{self.speed}")


    def center_map(self,event=None):
        self.map_widget.fit_bounding_box(self.topleft, self.bottomright)


    def refresh_database(self, event=None):
        self.database.refresh_dates(table="PATHS")
        self.cal.unhighlight()
        for _date in self.database.date_df:
            dat=date.fromisoformat(_date)
            self.cal.add_to_data(dat)
        self.refreshed = True

    def download_database(self, event=None):
        if self.play_status:
            self.play()
        if self.sm_state == 0: #Query Path
            if not self.busy:
                self.timeline.config(to=0)
                self.playbuttom["state"] = "disabled"
                self.date=self.cal.selection_get()
                self.date_var.set(self.date.isoformat())
                self.status_label.set("Querying Database")
                date_=self.date.strftime("%Y-%m-%d")
                query=f"SELECT Date,Latitude,Longitude,Heading,ASV FROM wqp.PATHS where date(Date) = '{date_}' AND Latitude<>0;"
                self.database.query(query=query)
                self.busy=True
            else:
                if not self.database.busy:
                    self.busy=False
                    self.path_db=self.database.df
                    self.sm_state=1
            self.download_loop_object = self.parent.after(20,self.download_database)

        if self.sm_state == 1: #Query Path
            if not self.busy:
                self.status_label.set("Querying WQP Database")
                date_=self.date.strftime("%Y-%m-%d")
                query=f"SELECT Date,Latitude,Longitude,Sensor,Data FROM wqp.WQP where date(Date) = '{date_}' AND Latitude<>0;"
                self.database.query(query=query)
                self.busy=True
            else:
                if not self.database.busy:
                    self.busy=False
                    self.wqp_db=self.database.df
                    self.sm_state=2
            self.download_loop_object = self.parent.after(20,self.download_database)

        elif self.sm_state == 2: #query trash
            if not self.busy:
                self.status_label.set("Querying Trash locations")
                date_=self.date.strftime("%Y-%m-%d")
                query=f"SELECT Date, Latitude_Obj, Longitude_Obj, ASV FROM wqp.OBJECT_POSITION where date(Date) = '{date_}' AND Latitude_Obj<>0;"
                self.database.query(query=query)
                self.busy=True
                self.download_loop_object = self.parent.after(20,self.download_database)

            else:
                if self.database.busy:
                    self.download_loop_object = self.parent.after(20,self.download_database)
                else:
                    #load database
                    self.trash_db=self.database.df
                    self.busy=False
                    self.sm_state=0
                    self.parent.after_cancel(self.download_loop_object)
                    self.load_database()


    def load_database(self):
        self.status_label.set("Processing Data")

        #take mean and only 1sec per data
        self.path_db = self.path_db.groupby(["Date"]).mean()
        if len(self.path_db) >3600:
            duration= "{:02d}".format(len(self.path_db)//3600) + ":" +"{:02d}".format((len(self.path_db)%3600)//60) + ":" + "{:02d}".format(len(self.path_db)%60)
        else:
            duration= "{:02d}".format((len(self.path_db)%3600)//60) + ":" + "{:02d}".format(len(self.path_db)%60)
        self.data_points_label.set(duration)
        if len(self.wqp_db) > 0:
            self.wqp_db = self.wqp_db.groupby(["Date","Sensor"]).mean().reset_index()
            self.wqp_db = self.wqp_db.set_index(["Date"])
            self.wqp_db = self.wqp_db[self.wqp_db["Data"]!=0]
        self.wqp_points_label.set(len(self.wqp_db))
        if len(self.trash_db) >0:
            self.trash_db = self.trash_db.groupby(["Date"]).mean()
        self.trash_points_label.set(len(self.trash_db))

        self.timeline.config(to=len(self.path_db))
        self.path_map = PATH(self.wqp_db, parent =self)
        for check in self.checkboxes:
            check["state"] = "normal"
        self.draw_map(0)
        self.go_to_time(0)
        self.playbuttom["state"] = "normal"
        self.status_label.set("Paused")



            
    def draw_map(self, event=None):
        for i in self.checkboxes:
            i.uncheck()
        self.checkboxes[event].check()
        if self.path_map is not None:
            self.path_map.select_sensor(self.database.sensors[event])


    def close(self):
        # stop play loop if running
        try:
            if self.play_status:
                self.play()
        except Exception:
            pass
        # cancel any pending download loop
        try:
            if hasattr(self, 'download_loop_object') and self.download_loop_object:
                self.parent.after_cancel(self.download_loop_object)
                self.download_loop_object = None
        except Exception:
            pass
        # attempt to destroy map widget and its markers to prevent callbacks after destruction
        try:
            if hasattr(self, 'map_widget') and self.map_widget is not None:
                try:
                    self.map_widget.delete_all_marker()
                except Exception:
                    pass
                try:
                    self.map_widget.destroy()
                except Exception:
                    pass
                self.map_widget = None
        except Exception:
            pass

    def open(self):
        if not self.refreshed:
            self.refresh_database()



#! /usr/bin/env python3
import re

import numpy as np

def extract_data(file):
    """
    @brief: This function takes in a .obj file and returns a list containing
            the indexes of the vertexes that make up the face, and a dict
            that contains the index of the vertex, along with its coordinates

    @param: .obj file

    @ret  : verticies (the dict containing coordinates)
    @ret  : faces (the list containing the vertexes' indexes)
    """

    verticies = np.empty((1, 3))
    faces = []

    # Read more about how waveform (.obj) files are structured to understand
    # how this code exactly works, but shortly:
    #   * If the line starts with a "v", then that's a vertex and what follows is
    #     its X, Y, Z coordinates (foats)
    #   * If the line starts with a "f", then that's a face and what follows is
    #     the list of verticies to be connected to create a face
    #     (formatted a bit strangely though, I recommend checking an example)

    for line in file.readlines():
        if line[0:2] == "v ":
            pt = np.array([[float(x) for x in re.findall("[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", line)]])
            verticies = np.concatenate([verticies, pt])
        elif line[0:2] == "f ":
            faces.append([int(vertex.split("/")[0]) for vertex in line[2:-2].split(' ')])
    return verticies, faces


if __name__ == '__main__':
    print("This is not the executable file, go to the 'main.py' file and run it instead!")



from aceti.map_utils import import_map, MapDownloader
import numpy as np
import matplotlib.pyplot as plt
import PIL

def plot_grid(map):
    ''' Plot a grid on top of a map. '''

    #check map properties
    if isinstance(map, str):
        base_matrix, lat_lon_map  = import_map(map)
    elif isinstance(map, tuple):
        base_matrix, lat_lon_map  = map
        if isinstance(base_matrix, np.ndarray) and isinstance(lat_lon_map, np.ndarray):
            if not np.all(np.isin(base_matrix, [0, 1])):
                raise ValueError("base_matrix must be a numpy array of 0s and 1s")
            if base_matrix.shape != lat_lon_map.shape:
                raise ValueError("base_matrix and lat_lon_map must have the same shape")
            if not np.all(np.apply_along_axis(lambda x: isinstance(x, tuple) and len(x) == 2 and all(isinstance(i, float) for i in x), 1, lat_lon_map)):
                raise ValueError("lat_lon_map must be a numpy array of tuples of 2 coordinates, [lat, lon]")
        else:
            raise ValueError("map must be a string or a tuple of [base_matrix, lat_lon_map]")
    else:
        raise ValueError("map must be a string or a tuple of [base_matrix, lat_lon_map]")
    # Plot the map
    print("lat_lon_map shape: ", lat_lon_map.shape)
    print("base_matrix shape: ", base_matrix.shape)
    print("lat_lon_map: ", lat_lon_map[0, 0])
    print("lat_lon_map: ", lat_lon_map[-1, -1])
    #to tuple
    northwest = [float(lat_lon_map[0, 0][0]), float(lat_lon_map[0, 0][1])]
    southeast = [float(lat_lon_map[-1, -1][0]), float(lat_lon_map[-1, -1][1])]
    map=MapDownloader(northwest, southeast, zoom=19, layer="s", server ="arcgis") 
    map.generateImage()
    # map.raw_map_img = PIL.Image.open("AlamilloPlantilla.png")
    # plt.imshow(map.map_img)
    # plt.savefig("alamillo95x216.png", dpi=600)
    map.show_grid(lat_lon_map, base_matrix)

    # plt.figure()
    # plt.imshow(base_matrix, cmap='gray', interpolation='nearest')
    # plt.show()

def save_as_csv():
    base_matrix, lat_lon_map  = import_map("/home/aloepacci/guadaira")

    # Ensure base_matrix is integer 0/1
    base_matrix = base_matrix.astype(int)

    # If lat_lon_map is an object array of tuples convert to numeric shape (nrows,ncols,2)
    # Normalize lat_lon_map to numeric shape (nrows, ncols, 2) with dtype float64
    if lat_lon_map.dtype == object:
        h, w = lat_lon_map.shape[:2]
        converted = np.zeros((h, w, 2), dtype=np.float64)
        for i in range(h):
            for j in range(w):
                v = lat_lon_map[i, j]
                if v is None:
                    converted[i, j, :] = np.nan
                else:
                    # Support (lat, lon) tuples or lists
                    converted[i, j, 0] = float(v[0])
                    converted[i, j, 1] = float(v[1])
        lat_lon_map = converted
    else:
        # if already numeric, ensure shape is (h,w,2)
        if lat_lon_map.ndim == 3 and lat_lon_map.shape[2] == 2:
            lat_lon_map = lat_lon_map.astype(np.float64)
        elif lat_lon_map.ndim == 2 and lat_lon_map.shape[1] == 2:
            # already flattened 2-col array, try to infer h,w from base_matrix
            try:
                h, w = base_matrix.shape
                lat_lon_map = lat_lon_map.reshape((h, w, 2)).astype(np.float64)
            except Exception:
                lat_lon_map = lat_lon_map.astype(np.float64)

    # Save base matrix as integer CSV
    np.savetxt("guadairaplantilla.csv", base_matrix, delimiter=" ", fmt='%d')

    # Save lat/lon grid with the same 2D shape as base_matrix: each cell is "lat,lon"
    h, w = base_matrix.shape
    if lat_lon_map.shape[0] != h or lat_lon_map.shape[1] != w:
        # try to reshape if lat_lon_map is flattened
        try:
            lat_lon_map = lat_lon_map.reshape((h, w, 2))
        except Exception:
            raise ValueError(f"lat_lon_map shape {lat_lon_map.shape} incompatible with base_matrix shape {base_matrix.shape}")

    str_grid = np.empty((h, w), dtype=object)
    for i in range(h):
        for j in range(w):
            lat = lat_lon_map[i, j, 0]
            lon = lat_lon_map[i, j, 1]
            # Format: lat,lon with 6 decimals
            str_grid[i, j] = f"{lat:.6f},{lon:.6f}"

    np.savetxt("guadairagrid.csv", str_grid, delimiter=";", fmt='%s')


# csv_to_npy("alamillo95x216")
# csv_to_png("alamillo95x216plantilla")

# plot_grid("alamilloaccess11x15")
# plot_grid("alamillo95x216")

# plot_grid("/home/aloepacci/guadaira")
# plot_grid("alamillo30x49")
# map=MapDownloader([37.191611287951396, -4.399097018031518], [37.188098777474906, -4.394347328263865], zoom=19, layer="s", server ="arcgis")
# map=MapDownloader([37.341664356846564, -5.858012026866486], [37.33741741302848, -5.855613536856963], zoom=19, layer="s", server ="arcgis")
# map.save_navigation_waters(name="guadaira")

# plt.imshow(map.map_img)
# plt.axis('off')
# plt.savefig("guadaira.png", dpi=600)


import matplotlib.pyplot as plt
base_matrix, lat_lon_map  = import_map("/home/aloepacci/trabajo/aceti/guadaira84")

base_matrix = base_matrix.T
plt.imshow(base_matrix, cmap='gray', interpolation='nearest')
plt.axis('off')
plt.gca().spines['top'].set_visible(False)
plt.gca().spines['right'].set_visible(False)
plt.gca().spines['left'].set_visible(False)
plt.gca().spines['bottom'].set_visible(False)
#quita todos los bordes en la imagen savefig
plt.savefig("guadaira_mask.png", dpi=600, bbox_inches='tight', pad_inches=0)
plt.show()


import time

def time_me(f):
	'''Decorator function to time functions' runtime in ms'''
	def wrapper(*args, **kwargs):
		start = time.time()
		res = f(*args, **kwargs)
		print(f'function: {f.__name__} took {(time.time()-start)*1000:.4f}ms')
		return res
	return wrapper


import tkintermapview
import os


# This scripts creates a database with offline tiles.

# specify the region to load (Hamburg)
top_left_position = (37.422587870100784, -6.003781676621391)
bottom_right_position = (37.41492544128397, -5.993943513241879)
zoom_min = 0
zoom_max = 22

# specify path and name of the database
script_directory = os.path.dirname(os.path.abspath(__file__))
database_path = os.path.join(script_directory, "assets/tiles/offline_tiles.db")

# create OfflineLoader instance
loader = tkintermapview.OfflineLoader(path=database_path,
                                      tile_server="https://mt0.google.com/vt/lyrs=s&hl=en&x={x}&y={y}&z={z}&s=Ga")

# save the tiles to the database, an existing database will extended
loader.save_offline_tiles(top_left_position, bottom_right_position, zoom_min, zoom_max)

# You can call save_offline_tiles() multiple times and load multiple regions into the database.
# You can also pass a tile_server argument to the OfflineLoader and specify the server to use.
# This server needs to be then also set for the TkinterMapView when the database is used.
# You can load tiles of multiple servers in the database. Which one then will be used depends on
# which server is specified for the TkinterMapView.

# print all regions that were loaded in the database
loader.print_loaded_sections()


import tkinter
import tkinter.font
import tkinter.ttk
import os, sys
import traceback
from PIL import Image, ImageTk
import requests
from io import BytesIO
import threading


class GIF():
    def __init__(self, file, speed=None, size = 0, master=None):
        self.path=file
        self.loc=0
        self.frames=[]
        self._master = master
        threading.Timer(0,self.__read__,args=(speed,size)).start()
    
    # corrected signature
    def frame(self, index):
        return self.frames[index]

    def get_next_frame(self):
        self.loc=(self.loc+1)%self.len
        return self.frames[self.loc]

    def __read__(self, speed, size):
        self.image=Image.open(self.path)
        if size == 0:
            try:
                n_frames=self.image.n_frames
            except EOFError:
                error= traceback.format_exc()
                print(f"error importing asset gif {self.path}:\n {error}")
                return

        for i in range(n_frames):
            # attach PhotoImage to a master (Tk root) to ensure proper tkinter lifecycle
            m = getattr(self, '_master', None) or tkinter._default_root
            self.frames.append(ImageTk.PhotoImage(self.image.copy(), master=m))
            self.image.seek(i)
        if speed is not None:
            self.delay=1000/speed
        else:
            try:
                self.delay = self.info.duration
            except:
                self.delay = 100

        if len(self.frames)==0:
            return
        self.len=len(self.frames)
        # print(f"gif {path.split('/')[-1]} len {self.len} delay {self.delay}")



class Assets(object):
    def __init__(self, path=None, master=None):
        # master should be the Tk root so PhotoImage objects are attached to it
        self._master = master or tkinter._default_root
        self.path=self.resource_path("assets/")
        # load images (attached to master)
        self.aceti_icon = self.include_image("icon_aceti.png",40,40)
        self.icon_water_trash = self.include_image("icon_water_trash.png",40,40)
        self.icon_trash_log = self.include_image("icon_trash_log.png",40,40)
        self.icon_play = self.include_image("play.png",35,35)
        self.icon_pause = self.include_image("pause.png",35,35)
        self.icon_ship = self.include_image("icon_ship.png",40,40)
        self.icon_gaussian_process = self.include_image("icon_gaussian.png",40,40)
        self.icon_on = self.include_image("on.png",40,20)
        self.icon_off = self.include_image("off.png",40,20)
        self.icon_path_planning = self.include_image("icon_path_planning.png",40,40)
        self.icon_test_big = self.include_image("icon_gaussian.png",400,400)
        self.icon_map_gen = self.include_image("mapgeneration.png",400,400)
        self.icon_lupa = self.include_image("lupa.png",220,220)
        self.map_tiles= self.resource_path("assets/tiles/offline_tiles.db")

        self.image_ship = Image.open(self.path+"icon_ship.png").convert('RGBA').resize((60,60))
        self.waste_len=2
        self.icon_waste=[]
        for i in range(self.waste_len):
            self.icon_waste.append(self.include_image(f"waste{i}.png",40,40))

        self.icon_measure=[]
        for i in range(11):
            self.icon_measure.append(self.include_image(f"measures/{i}.png",10,10))

    def include_image(self,image,sizex,sizey):
        try:
            originalImg = Image.open(self.path+image)
            img= originalImg.resize((sizex, sizey))
            # attach PhotoImage to the supplied master (Tk root) to avoid
            # lifecycle issues during interpreter shutdown
            m = getattr(self, '_master', None) or tkinter._default_root
            return ImageTk.PhotoImage(img, master=m)
        except:
            error = traceback.format_exc()
            print(f"error importing asset {image}:\n {error}")


    def resource_path(self, relative_path):
        """ Get absolute path to resource, works for dev and for PyInstaller """
        try:
            # PyInstaller creates a temp folder and stores path in _MEIPASS
            base_path = sys._MEIPASS
        except Exception:
            base_path = os.path.abspath(".")

        path = os.path.join(base_path, relative_path)
        if os.path.exists(path):
            return path
        else:#get path of the file
            base_path =  os.path.dirname(__file__)
            path = os.path.join(base_path, "assets")

        if not os.path.exists(base_path):
            raise FileNotFoundError(f"Asset path not found: {base_path}")

        return os.path.join(base_path, relative_path)
    
    def download_image(self, image):
        baseurl="https://fontawesome.com/start"
        response = requests.get(baseurl+"/"+image)
        img = Image.open(BytesIO(response.content))


from gui import GUI

import gc

class main_:
    ''' Class that launches everything '''

    def __init__(self, *args, **kwargs):
        self._initialise_gui(*args, **kwargs)
        self._update_display()
        self._gui.mainloop()

    def _initialise_gui(self, *args, **kwargs):
        self._gui = GUI(*args, **kwargs)

    def _update_display(self):
        self._gui.render()
        self._gui.after(1, self._update_display)

if __name__ == '__main__':
    main_()

gc.collect()


import tkinter
import tkinter.font
import tkinter.ttk
import os, sys, webbrowser, platform
import signal
import random
from tkintermapview import TkinterMapView
from tkintermapview.canvas_position_marker import CanvasPositionMarker
import threading
import time
import screeninfo
import traceback
import re
import requests
import math
import numpy as np
import pandas as pd
from PIL import Image, ImageTk
sys.path.append('../')
from ..shared import *
from ..assets import *
from ..widgets.ListboxEditable import ListboxEditable
from ..widgets.labels import *
from .map_creation_window import MapCreationWindow
max_retain_data=40


class MAP_GENERATOR_WIDGET(TkinterMapView):
    def __init__(self, *args, parent=None, **kwargs):
        self.assets=parent.assets
        self.parent=parent
        super().__init__(*args,**kwargs)


class MAP_GENERATORTAB(tkinter.ttk.PanedWindow):
    def __init__(self, parent=None):
        if parent is None:
            parent=tkinter.Tk()
        self.parent=parent
        self.polygon=None
        #inherit
        try:
            self.assets=self.parent.assets
            self.shared=self.parent.shared
        except:
            self.assets=Assets()
            self.shared=SHARED()

        super().__init__(orient="horizontal")
        self.gs_tab()

    def gs_tab(self):
        self.GPS_panel = tkinter.ttk.PanedWindow(orient="vertical")
        self.add(self.GPS_panel)

        #create map
        self.map_widget = MAP_GENERATOR_WIDGET(parent= self,corner_radius=0,width=int(self.parent.screenwidth*0.75)+1, height=int(self.parent.screenheight*0.75)+1) #this widget has no automatic size
        self.map_widget.set_tile_server("https://mt0.google.com/vt/lyrs=s&hl=en&x={x}&y={y}&z={z}&s=Ga", max_zoom=22)
        self.GPS_panel.add(self.map_widget)
        # self.map_widget.pack_propagate(False)
        self.topleft=[37.420088,-6.001346]
        self.bottomright=[37.418399,-5.997504]
        self.map_widget.fit_bounding_box(self.topleft, self.bottomright)

        #map selection
        self.map_selection=tkinter.Frame(self, width=20)
        self.map_selection.pack_propagate(False)
        self.add(self.map_selection)

        self.map_selection_frame=tkinter.Frame(self.map_selection)
        self.map_selection_frame.pack(side="top", fill="x")
        #Label barra de búsqueda
        self.map_search_label=tkinter.Label(self.map_selection_frame, text="1 Barra de Búsqueda",font=tkinter.font.Font(weight='bold', size=18))
        self.map_search_label.pack(side="top",pady=15)
        # search bar for tkintermapview
        self.search_frame = tkinter.Frame(self.map_selection_frame)
        self.search_entry = tkinter.Entry(self.search_frame)
        self.search_entry.pack(side="left", fill="x", expand=True, padx=(0,4))
        self.search_button = tkinter.Button(self.search_frame, image=self.assets.icon_lupa, command=self.search_location, width=36, height=36)
        self.search_button.pack(side="right")
        self.search_entry.bind("<Return>", self.search_location)
        self.search_frame.pack(side="top", fill="x", pady=(10,5))

        #L selecciona las esquinas del mapa
        self.map_search_label=tkinter.Label(self.map_selection_frame, text="2 Corner selector",font=tkinter.font.Font(weight='bold', size=18))
        self.map_search_label.pack(side="top",pady=15)


        # 2 botones para elegir puntos de coordenadas de dos esquinas cuando pulse en el mapa despues de estos
        self.northwest_selection_frame=tkinter.Frame(self.map_selection_frame)
        self.northwest_selection_frame.pack(side="top", fill="x", pady=(10,5))
        self.map_selection_button1 = tkinter.Button(self.northwest_selection_frame, text="Esquina 1", command=self.select_top_left)
        self.map_selection_button1.pack(side="left", pady=5)
        # use StringVar so we can trace changes
        self.northwest_var_lat = tkinter.StringVar()
        self.northwest_var_lon = tkinter.StringVar()
        self.northwest_entry_lat = tkinter.Entry(self.northwest_selection_frame, textvariable=self.northwest_var_lat)
        self.northwest_entry_lat.pack(side="left", fill="x", expand=True, padx=(0,4))
        self.northwest_entry_lon = tkinter.Entry(self.northwest_selection_frame, textvariable=self.northwest_var_lon)
        self.northwest_entry_lon.pack(side="left", fill="x", expand=True, padx=(0,4))

        self.southeast_selection_frame=tkinter.Frame(self.map_selection_frame)
        self.southeast_selection_frame.pack(side="top", fill="x", pady=(10,5))
        self.map_selection_button2 = tkinter.Button(self.southeast_selection_frame, text="Esquina 2", command=self.select_bottom_right)
        self.map_selection_button2.pack(side="left", pady=5)
        # use StringVar so we can trace changes
        self.southeast_var_lat = tkinter.StringVar()
        self.southeast_var_lon = tkinter.StringVar()
        self.southeast_entry_lat = tkinter.Entry(self.southeast_selection_frame, textvariable=self.southeast_var_lat)
        self.southeast_entry_lat.pack(side="left", fill="x", expand=True, padx=(0,4))
        self.southeast_entry_lon = tkinter.Entry(self.southeast_selection_frame, textvariable=self.southeast_var_lon)
        self.southeast_entry_lon.pack(side="left", fill="x", expand=True, padx=(0,4))

        # add traces so any change to these variables updates the polygon
        try:
            for var in (self.northwest_var_lat, self.northwest_var_lon, self.southeast_var_lat, self.southeast_var_lon):
                # trace_add is preferred in modern tkinter
                try:
                    var.trace_add('write', self.writepolygon)
                except Exception:
                    # fallback to older trace
                    var.trace('w', lambda *args, v=var: self.writepolygon())
        except Exception:
            pass


        #L selecciona las esquinas del mapa
        self.map_config_label=tkinter.Label(self.map_selection_frame, text="3 Configuracion",font=tkinter.font.Font(weight='bold', size=18))
        self.map_config_label.pack(side="top",pady=15)

        # Configuración de la selección del mapa
        self.map_config_frame = tkinter.Frame(self.map_selection_frame)
        self.map_config_frame.pack(side="top", fill="x", pady=(10, 5))

        tkinter.Label(self.map_config_frame, text="Grid Size (meters):").pack(side="left", padx=(0, 10))

        self.gridsize_var = tkinter.StringVar()
        self.gridsize_var.set("1.0")
        vcmd = (self.register(self._validate_float), '%P')
        self.grid_check = tkinter.Entry(self.map_config_frame, textvariable=self.gridsize_var)
        self.grid_check.pack(side="left", padx=(0, 10))
        self.grid_check.config(validate='key', validatecommand=vcmd)
        self.gridsize_var.trace_add('write', lambda *args: print(f"Grid size set to: {self.gridsize_var.get()}"))


        #botón de continuar
        self.map_selection_button = tkinter.Button(self.map_config_frame, text="Continue", command = self.to_map_creator)
        self.map_selection_button.pack(side="top", pady = 50)


        self.mouse_position = tkinter.StringVar()
        self.mouse_position.set("Lat: --, Lon: --")
        self.mouse_latlon=tkinter.Label(self.map_selection_frame , textvariable=self.mouse_position, width=int(self.parent.screenwidth*0.12), height=1, font=tkinter.font.Font(weight='bold', size=12))
        self.mouse_latlon.pack(side="top", padx=0, pady=2,anchor="c", fill="both")
        self.map_widget.canvas.bind("<Motion>", self.update_mouse_position)
        self.map_widget.canvas.bind("<Leave>", self.clear_mouse_position)

        
        #botón de continuar
        self.map_load_button = tkinter.Button(self.map_selection_frame, text="Load Previous map", command = self.load_previous_map)
        self.map_load_button.pack(side="top", pady = 50)




    def update_mouse_position(self, event=None):
        try:
            latitude, longitude = self.map_widget.convert_canvas_coords_to_decimal_coords(event.x, event.y)
            self.mouse_position.set(f"Lat: {latitude:.6f}, Lon: {longitude:.6f}")
        except Exception:
            pass

    def clear_mouse_position(self, event=None):
        self.mouse_position.set("Lat: --, Lon: --")


    def select_top_left(self):
        def callback(event):
            try:
                lat, lon = self.map_widget.convert_canvas_coords_to_decimal_coords(event.x, event.y)
            except Exception:
                # fallback: use event coordinates if conversion not available
                try:
                    lat, lon = event.lat, event.lon
                except Exception:
                    print("Could not get coordinates from click event")
                    return
            # write coordinates to entries
            try:
                # update traced variables (these will trigger writepolygon)
                try:
                    self.northwest_var_lat.set(f"{lat:.6f}")
                    self.northwest_var_lon.set(f"{lon:.6f}")
                except Exception:
                    # fallback: write directly to entries
                    self.northwest_entry_lat.delete(0, 'end')
                    self.northwest_entry_lat.insert(0, f"{lat:.6f}")
                    self.northwest_entry_lon.delete(0, 'end')
                    self.northwest_entry_lon.insert(0, f"{lon:.6f}")
            except Exception:
                pass
            # unbind the click handler
            try:
                self.map_widget.canvas.unbind("<Button-1>")
                self.map_widget.canvas.bind("<Button-1>", self.map_widget.mouse_click)

            except Exception:
                pass
        self.map_widget.canvas.unbind("<Button-1>")
        self.map_widget.canvas.bind("<Button-1>", callback)

    def _validate_float(self, proposed):
        """Return True if proposed string is empty or a float with up to 2 decimals."""
        if proposed == "":
            return True
        try:
            # allow a leading minus and decimals
            # but enforce at most 2 decimals while typing
            m = re.match(r'^-?\d*(?:\.(\d{0,2})?)?$', proposed)
            return m is not None
        except Exception:
            return False

    def select_bottom_right(self):
        def callback(event):
            try:
                lat, lon = self.map_widget.convert_canvas_coords_to_decimal_coords(event.x, event.y)
            except Exception:
                try:
                    lat, lon = event.lat, event.lon
                except Exception:
                    print("Could not get coordinates from click event")
                    return
            try:
                # update traced variables (these will trigger writepolygon)
                try:
                    self.southeast_var_lat.set(f"{lat:.6f}")
                    self.southeast_var_lon.set(f"{lon:.6f}")
                except Exception:
                    self.southeast_entry_lat.delete(0, 'end')
                    self.southeast_entry_lat.insert(0, f"{lat:.6f}")
                    self.southeast_entry_lon.delete(0, 'end')
                    self.southeast_entry_lon.insert(0, f"{lon:.6f}")
            except Exception:
                pass
            try:
                self.map_widget.canvas.unbind("<Button-1>")
                self.map_widget.canvas.bind("<Button-1>", self.map_widget.mouse_click)
            except Exception:
                pass
            print(f"Bottom-right corner set to: {self.southeast_entry_lat.get()}, {self.southeast_entry_lon.get()}")

        self.map_widget.canvas.unbind("<Button-1>")
        self.map_widget.canvas.bind("<Button-1>", callback)

    def search_location(self, event=None):
        try:
            query = self.search_entry.get().strip()
            if not query:
                return
            # simple rate-limit to avoid rapid repeated queries
            try:
                import time
                now = time.time()
                if getattr(self, '_last_search', 0) and now - self._last_search < 1.0:
                    # ignore searches less than 1s apart
                    return
                self._last_search = now
            except Exception:
                pass

            # First, try a direct Nominatim HTTP request with proper identification
            try:
                nominatim_url = "https://nominatim.openstreetmap.org/search"
                headers = {"User-Agent": "aceti_maps/0.0.2 (acasado4@us.es)"}
                params = {"q": query, "format": "jsonv2", "addressdetails": 1, "limit": 1, "email": "acasado4@us.es"}
                resp = requests.get(nominatim_url, params=params, headers=headers, timeout=10)
                if resp.status_code == 200:
                    data = resp.json()
                    if data:
                        lat = float(data[0]["lat"])
                        lon = float(data[0]["lon"])
                        try:
                            self.map_widget.set_position(lat, lon)
                            self.map_widget.set_zoom(19)
                            return
                        except Exception:
                            pass
                    else:
                        print(f"No geocoding results for '{query}'")
                else:
                    # Nominatim may return 403 if our requests are blocked/rate-limited
                    print(f"Nominatim HTTP error {resp.status_code} for query '{query}'")
                    if resp.status_code == 403:
                        print("Nominatim returned 403 — you may be rate-limited or blocked. Consider using an API key provider or running a local Nominatim instance.")
            except Exception as e:
                print(f"Direct Nominatim request failed: {e}")

            # Fallback to internal geocoder provided by tkintermapview (may hit same service)
            try:
                self.map_widget.set_address(query)
                return
            except Exception:
                # if that fails, try interpreting the query as 'lat, lon'
                m = re.match(r"\s*([+-]?\d+\.?\d*)[,\s]+([+-]?\d+\.?\d*)\s*$", query)
                if m:
                    lat = float(m.group(1))
                    lon = float(m.group(2))
                    try:
                        self.map_widget.set_position(lat, lon)
                        self.map_widget.set_zoom(15)
                        return
                    except Exception:
                        pass
                print(f"Search failed for '{query}' (no geocoding result)")
        except Exception as e:
            print(f"Error in search_location: {e}")


    def writepolygon(self, *args):
        """Draw or update a rectangular polygon from northwest and southeast variables."""
        try:
            lat1 = float(self.northwest_var_lat.get())
            lon1 = float(self.northwest_var_lon.get())
            lat2 = float(self.southeast_var_lat.get())
            lon2 = float(self.southeast_var_lon.get())
        except Exception:
            return

        # ensure lat1 is the northern (greater) and lat2 is southern (smaller) if user inverted
        try:
            if lat2 > lat1:
                lat1, lat2 = lat2, lat1
            if lon1 > lon2:
                lon1, lon2 = lon2, lon1
        except Exception:
            pass

        coords = [(lat1, lon1), (lat1, lon2), (lat2, lon2), (lat2, lon1), (lat1, lon1)]

        try:
            # remove previous polygon/path if exists
            try:
                if getattr(self, 'polygon', None):
                    try:
                        # some objects may provide a delete method
                        if hasattr(self.polygon, 'delete'):
                            self.polygon.delete()
                        elif hasattr(self.polygon, 'remove'):
                            self.polygon.remove()
                    except Exception:
                        pass
                    self.polygon = None
            except Exception:
                pass

            # try set_polygon (if available), otherwise use set_path
            try:
                self.polygon = self.map_widget.set_polygon(coords)
            except Exception:
                try:
                    # delete all paths to avoid duplicates
                    try:
                        self.map_widget.delete_all_path()
                    except Exception:
                        pass
                    self.polygon = self.map_widget.set_path(coords)
                except Exception as e:
                    print(f"Could not draw polygon: {e}")
        except Exception as e:
            print(f"writepolygon error: {e}")

    def center_map(self,event=None):
        self.map_widget.fit_bounding_box(self.topleft, self.bottomright)

    def to_map_creator(self, event=None):

        try:
            lat1 = float(self.northwest_var_lat.get())
            lon1 = float(self.northwest_var_lon.get())
            lat2 = float(self.southeast_var_lat.get())
            lon2 = float(self.southeast_var_lon.get())
            if lat2 > lat1:
                lat1, lat2 = lat2, lat1
            if lon1 > lon2:
                lon1, lon2 = lon2, lon1
            #if they look like numbers
            if not all(isinstance(coord, float) for coord in (lat1, lon1, lat2, lon2)):
                raise ValueError("Coordinates must be numeric.")
        except Exception as e:
            tkinter.messagebox.showerror("Invalid Coordinates", "Please enter valid numeric coordinates for both corners.")
            return
        try:
            MapCreationWindow(self, northwest=[lat1, lon1], southeast=[lat2, lon2], gridsize=float(self.gridsize_var.get()))
        except Exception as e:
            print(f"Error in to_map_creator: {e}")

    def load_previous_map(self, event=None):
        #open file selector
        file = tkinter.filedialog.askopenfilename(title="Select a map file", filetypes=[("Mapa", "*.csv *.npy")])
        try:
            from aceti.map_utils import import_map
            base_matrix, lat_lon_map = import_map(file)
            northwest = [lat_lon_map[0, 0, 0], lat_lon_map[0, 0, 1]]
            southeast = [lat_lon_map[-1, -1, 0], lat_lon_map[-1, -1, 1]]

            def estimate_gridsize(lat_lon_map):
                import pyproj
                northwest = [lat_lon_map[0, 0, 0], lat_lon_map[0, 0, 1]]
                southeast = [lat_lon_map[-1, -1, 0], lat_lon_map[-1, -1, 1]]
                to_proxy_transformer = pyproj.Transformer.from_crs('epsg:4326', 'epsg:3857')
                transformed_nw = to_proxy_transformer.transform(northwest[1], northwest[0]) # Transform NW point to 3857
                transformed_se = to_proxy_transformer.transform(southeast[1], southeast[0]) # .. same for SE
                width = int(abs(transformed_nw[1]-transformed_se[1]))
                height = int(abs(transformed_nw[0]-transformed_se[0]))
                cellwidth = width / base_matrix.shape[1]
                cellheight = height / base_matrix.shape[0]
                return round((cellwidth+cellheight)/2, 1)  # Assuming square cells, you could also average cellwidth and cellheight
            grid_size = estimate_gridsize(lat_lon_map)
        except Exception as e:
            tkinter.messagebox.showerror("Error Loading File", f"An error occurred while loading the map file: {e}")
            return
        try:
            MapCreationWindow(self, northwest=northwest, southeast=southeast, gridsize=grid_size, map_mask=base_matrix, latlon_map=lat_lon_map)
        except Exception as e:
            print(f"Error in load_previous_map: {e}")
    def close(self):
        pass

    def open(self):
        pass



import sys
sys.path.append('.')

import pandas as pd
import numpy as np
import os
from typing import Union
import lzma
import pickle
import yaml
from collections import defaultdict

class Legendarium:

    def __init__(self, experiment_name : str, experiment_description : str, path : str):

        self._exp_name = experiment_name
        self._exp_desc = experiment_description
        self._path = path

        # Check if directory exists
        if path is not None:
            if not os.path.exists(self._path):
                os.makedirs(self._path)
            else:
                # Check if the there is a file with the same experiment name
                if os.path.exists(os.path.join(self._path, f"{self._exp_name}.metrics.xz")):
                    raise Exception("Experiment already exists!")
        

        self.metrics_meta = {}
        self.metrics = []
        self.parameters = {}
        self.n_metrics = 0
        self.data_names = []

    def create_parameter(self, parameter_name : str, parameter_value):        
        """ 
        Create a parameter for the experiment
        """
        self.parameters.update({parameter_name : parameter_value})

    def create_metric(self, metric_name : str, data_type : type, description : str, unit : str):

        self.metrics_meta.update({metric_name : {"type" : data_type, "description" : description, "unit" : unit, "order": self.n_metrics}})
        self.n_metrics += 1

    def write(self, run : int, step : int, **kwargs):
        """ 
        Write data to the metric
        """

        # Check if metric exists
        for metric_name, new_data in kwargs.items():

            if metric_name not in self.metrics_meta:
                raise Exception(f"Metric {metric_name} not found!")
            
            # Check if the data type is correct
            if not isinstance(run, int):
                raise Exception("Run should be an integer!")
            
            if not isinstance(step, int):
                raise Exception("Step should be an integer!")
            
            # Check if the data type is correct
            if not issubclass(type(kwargs[metric_name]), self.metrics_meta[metric_name]["type"]):
                raise Exception(f"Data type mismatch for metric {metric_name}!")
            
        # Check if all the metrics are present
        if len(kwargs) != self.n_metrics:
            raise Exception("Not all metrics are present in the data!")
        
        new_data = [run, step] + [None] * self.n_metrics
    
        
        for key, value in kwargs.items():
            
            # Comprobamos si la clave es una metrica registrada
            if key not in self.metrics_meta:
                raise Exception(f"Metric {key} not found!")
            
            # Comprobamos si el tipo de dato es correcto
            if not isinstance(value, self.metrics_meta[key]["type"]):
                raise Exception(f"Data type mismatch for metric {key}!")
            
            # Obtenemos la posición en la lista
            pos = self.metrics_meta[key]["order"]
            new_data[pos + 2] = value

        # Añadimos la información de los parámetros
        for key, value in self.parameters.items():
            new_data.append(value)
        
        self.metrics.append(new_data)


    def save(self):
        """
        Save the experiment
        """

        data_names = ["run", "step"] + list(self.metrics_meta.keys()) + list(self.parameters.keys())


        # Save the metrics
        with lzma.open(os.path.join(self._path, f"{self._exp_name}.metrics.xz"), "wb") as f:
            pickle.dump(self.metrics, f)

        # Save the metrics meta
        with open(os.path.join(self._path, f"{self._exp_name}.meta.yaml"), "w") as f:
            yaml.dump(data_names, f)


def load_experiment_pd(experiment_name : str, path : str):
    # Load the metrics as a pandas dataframe

    # Load the metrics
    with lzma.open(os.path.join(path, f"{experiment_name}.metrics.xz"), "rb") as f:
        metrics = pickle.load(f)

    # Load the metrics meta
    with open(os.path.join(path, f"{experiment_name}.meta.yaml"), "r") as f:
        data_names = yaml.load(f, Loader=yaml.FullLoader)

    # Create the dataframe
    df = pd.DataFrame(metrics, columns = data_names)

    return df

def load_experiments(path : str):
    # Load all the experiments in a directory and return a pandas dataframe

    pds = []

    for file in os.listdir(path):
        if file.endswith(".metrics.xz"):
            experiment_name = file.split(".")[0]
            try:
                df = load_experiment_pd(experiment_name, path)
                pds.append(df)
            except Exception as e:
                print(f"Error loading experiment {experiment_name}: {e}")


    return pd.concat(pds)


if __name__ == "__main__":


    # Create an instance of the class
    exp = Legendarium(f"test", "Test experiment", "experiments")

    # Create a parameter
    exp.create_parameter("algorithm", "Greedy")
    exp.create_parameter("max_distance", 100.0)

    # Create a metric
    exp.create_metric("reward", float, "Reward function", "points")
    exp.create_metric("map", np.ndarray, "Mean map", "%")

    for run in range(3):

        # Write data
        for i in range(100):
            exp.write(run = run, step = i, reward = np.random.rand(), map = np.random.rand(100,100))

    # Save the data
    exp.save()

    # Load the data
    df = load_experiments("experiments")
    print(df.head())

    import seaborn as sns
    import matplotlib.pyplot as plt

    sns.lineplot(data = df, x = "step", y = "reward", hue = "run")
    plt.show()


import tkinter
import tkinter.font
import tkinter.ttk

class Checkbox(tkinter.ttk.Checkbutton):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.variable = tkinter.BooleanVar(self)
        self.configure(variable=self.variable)
    
    def checked(self):
        return self.variable.get()
    
    def check(self):
        self.variable.set(True)
    
    def uncheck(self):
        self.variable.set(False)

    def mouse_enter(self,event=None):
        self.set_text(self.name)
        super().mouse_enter(event)
        self.draw()

    def mouse_leave(self, event=None):
        self.set_text("")
        super().mouse_leave(event)

    def leftclick(self, event=None):
        pass




# modified from the scripts created by Hayden Eskriett [http://eskriett.com] and jbndlr
# Find the associated blog post at: http://blog.eskriett.com/2013/07/19/downloading-google-maps/
# but it has been done in a better way using the data in https://developers.google.com/maps/documentation/javascript/coordinates
# and https://stackoverflow.com/questions/40342355/how-can-i-generate-a-regular-geographic-grid-using-python





###################################IMPORTANT##################################
#Google maps works with [Latitude Longitude] variables however
#Dronekit and other programs (such as gridder) work with [Longitude Latitude] variables
##############################################################################

import urllib.request
from PIL import Image
import os
import math
import pyproj
import matplotlib.pyplot as plt
import numpy as np
# from KMLMissionGeneration import KMLMissionGenerator
class CustomError(Exception):
     pass

class GoogleMapDownloader:
    """
        A class which generates high resolution google maps image given
        Northwest and Southeast points in [Lat Lon] and zoom level
        You can change the quality of the image by changing the zoom and layer values
        
        19 is a valid zoom value knowing the size of our drone
        
        for the maps you have available
        ROADMAP = "v"
        TERRAIN = "p"
        ALTERED_ROADMAP = "r"
        SATELLITE = "s"
        TERRAIN_ONLY = "t"
        HYBRID = "y"
        
        
    """

    def __init__(self, Northwest, Southeast, zoom=19, layer="s"):
        """
            GoogleMapDownloader Constructor
            Args:
                lat:    The latitude of the location required
                lng:    The longitude of the location required
                zoom:   The zoom level of the location required, ranges from 0 - 23
        """
        self.Northwest = Northwest
        self.Southeast = Southeast
        self._zoom = zoom
        self._layer = layer
        
        if Northwest[0]<=Southeast[0] or Northwest[1]>=Southeast[1]:
            raise CustomError("You must use Northwest(top-left) and Southeast(bottom-right) coordinates")
        self.get_nw_coordinates()
        self.get_se_coordinates()
        
        if (abs(self.nw_tile[0]-self.se_tile[0]) == 0) or (abs(self.nw_tile[1]-self.se_tile[1]) == 0):
            raise CustomError("Insuficient zoom, points are too close, i am lazy to program this, increase zoom")
        
        

    def get_nw_coordinates(self):
        """
            Generates an X,Y tile and pixel coordinate based on the latitude, longitude
            and zoom level
            Returns:    An X,Y pixel coordinate
        """

        tile_size = 256 #each tile has 256 pixels

        # Use a left shift to get the power of 2, zoom 0 is a world map, zoom 1 is wold map divided in 4 tiles, zoom 2 is world map divided in 16 tiles
        # i.e. a zoom level of 2 will have 2^2 = 4 tiles as coordinate sistem divides axis in 4 (x+y+,x-y+,x-y-,x+y-)
        numTiles = 1 << self._zoom

        # Find the x_pixel given the longitude
        x_pixel = (tile_size / 2 + self.Northwest[1] * tile_size / 360.0) * numTiles 

        # Convert the latitude to radians and take the sine
        sin_y = math.sin(self.Northwest[0] * (math.pi / 180.0))

        # Calulate the y_pixel
        y_pixel = ((tile_size / 2) + 0.5 * math.log((1 + sin_y) / (1 - sin_y)) * -(
        tile_size / (2 * math.pi))) * numTiles 
        
        self.nw_pixel=[x_pixel, y_pixel]
        self.nw_tile=[int(x_pixel // tile_size), int(y_pixel // tile_size)]
    
    def get_se_coordinates(self):
        """
            Generates an X,Y tile and pixel coordinate based on the latitude, longitude
            and zoom level
            Returns:    An X,Y pixel coordinate
        """

        tile_size = 256 #each tile has 256 pixels

        # Use a left shift to get the power of 2, zoom 0 is a world map, zoom 1 is wold map divided in 4 tiles, zoom 2 is world map divided in 16 tiles
        # i.e. a zoom level of 2 will have 2^2 = 4 tiles as coordinate sistem divides axis in 4 (x+y+,x-y+,x-y-,x+y-)
        numTiles = 1 << self._zoom

        # Find the x_pixel given the longitude
        x_pixel = (tile_size / 2 + self.Southeast[1] * tile_size / 360.0) * numTiles 

        # Convert the latitude to radians and take the sine
        sin_y = math.sin(self.Southeast[0] * (math.pi / 180.0))

        # Calulate the y_pixel
        y_pixel = ((tile_size / 2) + 0.5 * math.log((1 + sin_y) / (1 - sin_y)) * -(
        tile_size / (2 * math.pi))) * numTiles 
        
        self.se_pixel=[x_pixel, y_pixel]
        self.se_tile=[math.ceil(x_pixel // tile_size)+1, math.ceil(y_pixel // tile_size)+1]
            
    def generateImage(self, ):
        """
            Generates an image by stitching a number of google map tiles together.

            Args:
                start_x:        The top-left x-tile coordinate
                start_y:        The top-left y-tile coordinate
                tile_width:     The number of tiles wide the image should be -
                                defaults to 5
                tile_height:    The number of tiles high the image should be -
                                defaults to 5
            Returns:
                A high-resolution Goole Map image.
        """
        tile_width = abs(self.nw_tile[0]-self.se_tile[0])
        tile_height = abs(self.nw_tile[1]-self.se_tile[1])
        
        
            
        # Determine the size of the image, for simplicity, we will print the whole image and later crop it
        width, height = 256 * tile_width, 256 * tile_height

        # Create a new image of the size required
        map_img = Image.new('RGB', (width, height))
        
        
        # calculate northwest pixel of the tile
        start_x=self.nw_tile[0]
        start_y=self.nw_tile[1]

        # print the map
        for x in range(0, tile_width):
            for y in range(0, tile_height):
                
                url = f'https://mt0.google.com/vt?lyrs={self._layer}&x=' + str(start_x + x) + '&y=' + str(start_y + y) + '&z=' + str(
                    self._zoom)

                current_tile = str(x) + '-' + str(y)
                urllib.request.urlretrieve(url, current_tile)

                im = Image.open(current_tile)
                map_img.paste(im, (x * 256, y * 256))

                os.remove(current_tile)
        
        
        # cut corners
        left=abs(self.nw_tile[0] * 256 - self.nw_pixel[0])
        top=abs(self.nw_tile[1] * 256 - self.nw_pixel[1])
        right=left + abs(self.se_pixel[0] - self.nw_pixel[0])
        bottom=top + abs(self.nw_pixel[1] - self.se_pixel[1])
        
        
        map_img = map_img.crop((left, top, right, bottom))
        return map_img

class gridder:
    
    def __init__(self, NW_corner, SE_corner, grid_size, zoom =19):
        """
            Args:
                NW_corner:
                SE_corner:   
                grid_size:   
        """
        
        
        
        # Set up transformers, EPSG:3857 is metric, same as EPSG:900913
        self.to_proxy_transformer = pyproj.Transformer.from_crs('epsg:4326', 'epsg:3857')
        self.to_original_transformer = pyproj.Transformer.from_crs('epsg:3857', 'epsg:4326')
        
        # Project corners to target projection
        self.NW_corner=[NW_corner[1],NW_corner[0]]
        self.SE_corner=[SE_corner[1],SE_corner[0]]
        
        #this transformation works in [Longitude Latitude]
        self.transformed_nw = self.to_proxy_transformer.transform(self.NW_corner[1], self.NW_corner[0]) # Transform NW point to 3857
        self.transformed_se = self.to_proxy_transformer.transform(self.SE_corner[1], self.SE_corner[0]) # .. same for SE
        
        self.pole=["E","N"]
        self.grid_size=grid_size
        self.zoom = zoom
        print(f"map size is {int(abs(self.transformed_nw[1]-self.transformed_se[1]))}x{int(abs(self.transformed_nw[0]-self.transformed_se[0]))}m")
        self.create_grid()
        print(f"grid size is {self.width}x{self.heigth} cells")
        self.calculate_pixels_per_grid()
        
        
        
        
        
    def save_to_file(self, name):
        with open(name, 'w') as of:
            for j in range(self.heigth):
                for i in range(self.width):
                    of.write('{:f},{:f};'.format(self.gridpoints[j][i][0], self.gridpoints[j][i][1]))
                of.write('\n')
            
            
    def create_grid(self):
        # Iterate over 2D area
        self.gridpoints = []
        countx=0
        county=0
        x = self.transformed_nw[0]
        if self.transformed_nw[0] >self.transformed_se[0]:
            self.pole[0]="N"
        else:
            self.pole[0]="S"
        while abs(y) > abs(self.transformed_se[1]):
            countx+=1
            county=0
            column=[]
            y = self.transformed_nw[0]
            if self.transformed_nw[0] <self.transformed_se[0]:
                self.pole[1]="E"
            else:
                self.pole[1]="W"
            while abs(x) > abs(self.transformed_se[0]):
                p = self.to_original_transformer.transform(x, y)
                column.append([p[0],p[1]])
                if self.pole[1]=="E":
                    x += self.grid_size
                else:
                    x -= self.grid_size
                countx+=1
            column.pop() #cut corners
            self.gridpoints.append(column)
            if self.pole[0]=="N":
                y -=self.grid_size
            else:
                y +=self.grid_size
        self.gridpoints.pop() #cut corners
        self.width=len(self.gridpoints[0])
        self.heigth=len(self.gridpoints)
            
        
    def to_pixel(self, GPS_point, zoom=19):
        """
            Generates an X,Y pixel coordinate based on the latitude, longitude
            and zoom level
            Returns:    An X,Y pixel coordinate
        """
        #check if we are inside the map
        if self.pole[1]=="W":
            if ((self.SE_corner[1])>(GPS_point[1])>(self.NW_corner[1])):
                raise CustomError(f"point {GPS_point} 1W is outside the map {[self.NW_corner,self.SE_corner]}")
        else:
            if ((self.SE_corner[1])<(GPS_point[1])<(self.NW_corner[1])):
                raise CustomError(f"point {GPS_point} 1E is outside the map {[self.NW_corner,self.SE_corner]}")
        if self.pole[0]=="N":
            if ((self.SE_corner[0])>(GPS_point[0])>(self.NW_corner[0])):
                raise CustomError(f"point {GPS_point} 0N is outside the map {[self.NW_corner,self.SE_corner]}")
        else:
            if ((self.SE_corner[0])<(GPS_point[0])<(self.NW_corner[0])):
                raise CustomError(f"point {GPS_point} 0S is outside the map {[self.NW_corner,self.SE_corner]}")



        tile_size = 256 #each tile has 256 pixels

        # Use a left shift to get the power of 2, zoom 0 is a world map, zoom 1 is wold map divided in 4 tiles, zoom 2 is world map divided in 16 tiles
        # i.e. a zoom level of 2 will have 2^2 = 4 tiles as coordinate sistem divides axis in 4 (x+y+,x-y+,x-y-,x+y-)
        numTiles = 1 << self.zoom

        # Find the x_pixel given the longitude
        x_pixel = (tile_size / 2 + GPS_point[1] * tile_size / 360.0) * numTiles 

        # Convert the latitude to radians and take the sine
        sin_y = math.sin(GPS_point[0] * (math.pi / 180.0))

        # Calulate the y_pixel
        y_pixel = ((tile_size / 2) + 0.5 * math.log((1 + sin_y) / (1 - sin_y)) * -(
        tile_size / (2 * math.pi))) * numTiles 
        
        return [x_pixel-self.corner_pixel[0], y_pixel-self.corner_pixel[1]]
    
    def calculate_pixels_per_grid(self):
        
        self.calculate_corner_pixel()
        p=self.to_pixel(self.gridpoints[self.heigth-1][self.width-1])
        x=p[0]
        y=p[1]
        self.pixel_width=x/(self.width-1)
        self.pixel_heigth=y/(self.heigth-1)
        print(f"pixel size is {x}x{y} with corner {self.corner_pixel}")
            
    def calculate_corner_pixel(self):
        """
            Generates an X,Y pixel coordinate based on the latitude, longitude
            and zoom level
            Returns:    An X,Y pixel coordinate
        """

        tile_size = 256 #each tile has 256 pixels

        # Use a left shift to get the power of 2, zoom 0 is a world map, zoom 1 is wold map divided in 4 tiles, zoom 2 is world map divided in 16 tiles
        # i.e. a zoom level of 2 will have 2^2 = 4 tiles as coordinate sistem divides axis in 4 (x+y+,x-y+,x-y-,x+y-)
        numTiles = 1 << self.zoom

        # Find the x_pixel given the longitude
        x_pixel = (tile_size / 2 + self.NW_corner[0] * tile_size / 360.0) * numTiles 

        # Convert the latitude to radians and take the sine
        sin_y = math.sin(self.NW_corner[1] * (math.pi / 180.0))

        # Calulate the y_pixel
        y_pixel = ((tile_size / 2) + 0.5 * math.log((1 + sin_y) / (1 - sin_y)) * -(
        tile_size / (2 * math.pi))) * numTiles 
        
        self.corner_pixel=[x_pixel, y_pixel]
        
    def show_grid(self, img):
            
        #show grid
        x=[]
        y=[]
        plt.figure()
        plt.imshow(img, cmap=plt.get_cmap('binary'))
        
        #horizontal lines
        for j in range(self.heigth):
            for i in range(self.width):
                p=self.to_pixel(self.gridpoints[j][i])
                x.append(p[0])
                y.append(p[1])
            plt.plot(x,y,'b-', linewidth=0.2)
            x=[]
            y=[]
        #vertical lines
        
        for i in range(self.width):
            for j in range(self.heigth):
                p=self.to_pixel(self.gridpoints[j][i])
                x.append(p[0])
                y.append(p[1])
            plt.plot(x,y,'b-', linewidth=0.2)
            x=[]
            y=[]
        plt.savefig('grid.png', dpi=400)
        
        img=Image.open('grid.png')
        img.show()
        
        
    def show_point(self, img, point):
        
        #print a point in the map
        pixel=self.to_pixel(point)
        
        plt.figure()
        plt.imshow(img, cmap=plt.get_cmap('binary'))
        plt.plot(pixel[0],pixel[1],'ro', markersize=5)
        plt.show()
        
    def save_empty_mask(self, file):
        with open(file, 'w') as of:
            for j in range(self.heigth-1):
                for i in range(self.width-1):
                    of.write('0;')
                of.write('\n')
                
    def save_mask(self, file):
        with open(file, 'w') as of:
            for j in range(self.heigth-1):
                for i in range(self.width-2):
                    of.write('{:d};'.format(self.mask[j][i]))
                of.write('{:d}\n'.format(self.mask[j][i]))
        
    def create_mask(self, img):
        grayImage=img.convert(mode="L")
        array = np.array(grayImage)
        self.mask=[]
        for j in range(self.heigth-1):
            column=[]
            for i in range(self.width-1):
                p1=self.to_pixel(self.gridpoints[j][i])
                p2=self.to_pixel(self.gridpoints[j+1][i])
                p3=self.to_pixel(self.gridpoints[j][i+1])
                p4=self.to_pixel(self.gridpoints[j+1][i+1])
                p1=array[int(p1[1])][int(p1[0])]
                p2=array[int(p2[1])][int(p2[0])]
                p3=array[int(p3[1])][int(p3[0])]
                p4=array[int(p4[1])][int(p4[0])]
                white=0
                if p1:
                    white+=1
                if p2:
                    white+=1
                if p3:
                    white+=1
                if p4:
                    white+=1
                if white>2:
                    column.append(0)
                else:
                    column.append(1)
            self.mask.append(column)
    def show_mask(self, img):
        #show grid
        
        plt.figure()
        plt.imshow(img, cmap=plt.get_cmap('binary'))
        
        for j in range(self.heigth-1):
            for i in range(self.width-1):
                if not self.mask[j][i]:
                    x=[]
                    y=[]
                    p1=self.to_pixel(self.gridpoints[j][i])
                    p2=self.to_pixel(self.gridpoints[j+1][i])
                    p3=self.to_pixel(self.gridpoints[j][i+1])
                    p4=self.to_pixel(self.gridpoints[j+1][i+1])
                    x.append(p1[0])
                    y.append(p1[1])
                    x.append(p2[0])
                    y.append(p2[1])
                    x.append(p4[0])
                    y.append(p4[1])
                    x.append(p3[0])
                    y.append(p3[1])
                    x.append(p1[0])
                    y.append(p1[1])
                    plt.plot(x,y,'b-', linewidth=0.2)
                    
        
        plt.savefig('mask.png', dpi=400)
        
        img=Image.open('mask.png')
        img.show()
        
    # def show_mission(self, img):
    #     plt.figure()
    #     plt.imshow(img, cmap=plt.get_cmap('binary'))
    #     point=0
    #     KML=KMLMissionGenerator("MisionesLoyola_dron_3.kml")
    #     mission=KML.get_mission_list()[0]
        
    #     for j in range(self.heigth-1):
    #         for i in range(self.width-1):
    #             if not self.mask[j][i]:
    #                 x=[]
    #                 y=[]
    #                 p1=self.to_pixel(self.gridpoints[j][i])
    #                 p2=self.to_pixel(self.gridpoints[j+1][i])
    #                 p3=self.to_pixel(self.gridpoints[j][i+1])
    #                 p4=self.to_pixel(self.gridpoints[j+1][i+1])
    #                 x.append(p1[0])
    #                 y.append(p1[1])
    #                 x.append(p2[0])
    #                 y.append(p2[1])
    #                 x.append(p4[0])
    #                 y.append(p4[1])
    #                 x.append(p3[0])
    #                 y.append(p3[1])
    #                 x.append(p1[0])
    #                 y.append(p1[1])
    #                 plt.plot(x,y,'b-', linewidth=0.2)
    #     for i in mission:
    #         point=[i[1], i[0]]
    #         pixel=self.to_pixel(point)
    #         plt.plot(pixel[0],pixel[1],'ro', markersize=0.5)
        
        
    #     plt.savefig('mission.png', dpi=400)
        
    #     img=Image.open('mission.png')
    #     img.show()
                    
                    
def main():
    # Create a new instance of GoogleMap Downloader
    """
    Loyola
    NW_corner=[37.30877610927852, -5.940516378146266] #northwest point of the map [latitude longitude]
    SE_corner=[37.30661986681041, -5.939422309058239] #southeast point of the map [latitude longitude]
    

    NW_corner=[36.123388953996404, -5.443591065870082] #northwest point of the map [latitude longitude]
    SE_corner=[36.11595558243919, -5.430975637057081] #southeast point of the map [latitude longitude]
    #alamillo

    NW_corner=[36.227154628215054, -6.188738765767977] #northwest point of the map [latitude longitude]
    SE_corner=[35.67592758704647, -5.13798693812742] #southeast point of the map [latitude longitude]
    #Estrecho
    """
    # NW_corner=[30.719200,-86.115300] #northwest point of the map [latitude longitude]
    # SE_corner=[30.7148,-86.110014] #southeast point of the map [latitude longitude]
    # #Lake DeFuniak
    
    # NW_corner=[37.420088017565696, -6.001345595400041] #northwest point of the map [latitude longitude]
    # SE_corner=[37.41837443521307, -5.997484063063451] #southeast point of the map [latitude longitude]
    #alamillo
    NW_corner=[36.57822688776558, -6.206246380974115] #northwest point of the map [latitude longitude]
    SE_corner=[36.577186700004944, -6.204225771439034] #southeast point of the map [latitude longitude]


    #SE_corner=[37.29989089621092, -5.9274044593699236]
    grid_size= 2.5     #size of the grid in meters
    zoom=19
    
    #create class of the map downloader 
    gmd = GoogleMapDownloader(NW_corner ,SE_corner, zoom =zoom) #zoom 19, aproximadamente tiles de 50m^2, más zoom no se aprecian obstaculos nuevos
    
    try:
        # Get the high resolution image
        img = gmd.generateImage()
    except IOError:
        print("Could not generate the image - try adjusting the zoom level and checking your coordinates")
    else:
        # Save the image to disk
        img.rotate(17)
        img.save("high_resolution_image2.png")
        print("The map has successfully been created")
        
    
    #Show the image
    plt.figure()
    plt.imshow(img, cmap=plt.get_cmap('binary'))
    plt.show()
    
    #create grid
    print(f"generating map for {NW_corner}, {SE_corner}")
    grid= gridder(NW_corner,SE_corner,grid_size, zoom =zoom)



    #show the grid on the image
    aux=0
    while aux!="y" and aux!="n":
        aux=input("is there already an existing mask (y/n)").lower()
    if aux=='n':
        grid.show_grid(img)
        grid.save_to_file("grid.csv")
        grid.save_empty_mask("empty_mask.csv")
        return
    
    while not os.path.exists(str("./"+aux)) or aux[-1]!="g":
        aux=input("insert mask name: ").lower()
    imgn=Image.open(aux)
    grid.show_grid(imgn)
    name=aux[0:-4]
    grid.create_mask(imgn)
    grid.save_mask(str(name+".csv"))
    grid.show_mask(img)


    #show a point
    #grid.show_point(img, [37.307659405593995, -5.940044140455814])
    
    #grid.show_mission(img)
        


if __name__ == '__main__':  main()


import obj_files_handler as obj_files_handler 
from geometry import Geometry
from timer import time_me

import math
import tkinter as tk
from tkinter import ttk, filedialog, colorchooser, messagebox
import pyscreenshot as ImageGrab

import sv_ttk

MIN_WIDTH = 1165
MIN_HEIGHT = 630

class GUI(tk.Tk):
	''''''
	CANVAS_WIDTH = 1000
	CANVAS_HEIGHT = 570
	CANVAS_COLOR = 'white'
	COMMON_X = 0.98	# Many graphical elements share the same relative X position
	MOVING_STEP = 10
	
	POINT_SIZE = 1 
	POINT_COLOR = '#131313'

	def __init__(self, title='3D_Viz', min_size=(MIN_WIDTH, MIN_HEIGHT)):
		''''''
		super().__init__()
		# Set the theme to be dark (there must be an initialized app)
		sv_ttk.set_theme("dark")
		
		self._file_exists = False # A flag for whether the file has been loaded or not
		self._changed = True # A flag used to only redraw the object when a change occured
		self._geometry_handler = Geometry(self.CANVAS_WIDTH, self.CANVAS_HEIGHT)
		
		# Initial canvas ratios
		self._canvas_w = int((self.CANVAS_WIDTH/MIN_WIDTH)*MIN_WIDTH)
		self._canvas_h = int((self.CANVAS_HEIGHT/MIN_HEIGHT)*MIN_HEIGHT)
		
		self._fill_color_holder = "#000000"
		self._line_color_holder = "#0000FF"

		self.__initialise_window(title, min_size)
		self.__create_widgets()
		self.__reset_rotation()

	def __initialise_window(self, title, min_size):
		self.title(title)
		self.minsize(*min_size)
		
	def __create_widgets(self):
		self.__create_canvas()
		self.__create_zoom_slider()
		self.__create_x_rot_slider()
		self.__create_y_rot_slider()
		self.__create_z_rot_slider()
		self.__create_reset_rot_button()
		self.__create_import_file_button()
		self.__create_screenshot_button()
		self.__create_up_down_left_right_buttons()
		self.__create_color_pickers()
		self.__create_fill_check()

	def __create_canvas(self):
		self._canvas_color = tk.StringVar()
		self._canvas_color.set("#FFFFFF")
		self._canvas = tk.Canvas(self, bg=self._canvas_color.get())
		self._canvas.place(relx=0.01, rely=0.01, relwidth=self.CANVAS_WIDTH/MIN_WIDTH, relheight=self.CANVAS_HEIGHT/MIN_HEIGHT)
		# Catch the canvas resize event
		self._canvas.bind("<Configure>", self.__resized)

		# Bind the scrolling on canvas event to zoom slider (Button-4/5 on linux)
		self._canvas.bind('<Button-4>', self.__mousewheel_scroll_in_canvas_up_event)
		self._canvas.bind('<Button-5>', self.__mousewheel_scroll_in_canvas_down_event)

	def __create_zoom_slider(self):
		ttk.Label(self, text="Zoom:").place(relx=self.COMMON_X, rely=0.052, relheight=0.035, relwidth=0.1, anchor="ne")
		self._zoom_slider = ttk.Scale(self, from_=1000.0, to=1.1, orient="horizontal", command=self.__changed)
		self._zoom_slider.set(self._geometry_handler.zoom)
		self._zoom_slider.place(relx=self.COMMON_X, rely=0.088, relheight=0.04, relwidth=0.1, anchor="ne")

	def __create_x_rot_slider(self):
		self._check_x_continuos = tk.BooleanVar()
		ttk.Checkbutton(self, text="", variable=self._check_x_continuos, command=self.__continuous_x, onvalue=True, offvalue=False).place(relx=self.COMMON_X-0.1, rely=0.121)
		ttk.Label(self, text="X Rotation:").place(relx=self.COMMON_X, rely=0.123, relheight=0.035, relwidth=0.075, anchor="ne")
		self.x_rotation_slider = ttk.Scale(self, from_=-math.pi, to=math.pi, orient="horizontal", command=self.__changed)
		self.x_rotation_slider.set(0)
		self.x_rotation_slider.place(relx=self.COMMON_X, rely=0.163, relheight=0.04, relwidth=0.1, anchor="ne")

	def __create_y_rot_slider(self):
		self._check_y_continuos = tk.BooleanVar()
		ttk.Checkbutton(self, text="", variable=self._check_y_continuos, command=self.__continuous_y, onvalue=True, offvalue=False).place(relx=self.COMMON_X-0.1, rely=0.202)
		ttk.Label(self, text="Y Rotation:").place(relx=self.COMMON_X, rely=0.204, relheight=0.035, relwidth=0.075, anchor="ne")
		self.y_rotation_slider = ttk.Scale(self, from_=-math.pi, to=math.pi, orient="horizontal", command=self.__changed)
		self.y_rotation_slider.set(0)
		self.y_rotation_slider.place(relx=self.COMMON_X, rely=0.244, relheight=0.04, relwidth=0.1, anchor="ne")

	def __create_z_rot_slider(self):
		self._check_z_continuos = tk.BooleanVar()
		ttk.Checkbutton(self, text="", variable=self._check_z_continuos, command=self.__continuous_z, onvalue=True, offvalue=False).place(relx=self.COMMON_X-0.1, rely=0.283)
		ttk.Label(self, text="Z Rotation:").place(relx=self.COMMON_X, rely=0.285, relheight=0.035, relwidth=0.075, anchor="ne")
		self.z_rotation_slider = ttk.Scale(self, from_=-math.pi, to=math.pi, orient="horizontal", command=self.__changed)
		self.z_rotation_slider.set(0)
		self.z_rotation_slider.place(relx=self.COMMON_X, rely=0.325, relheight=0.04, relwidth=0.1, anchor="ne")

	def __create_reset_rot_button(self):
		ttk.Button(self, text="Reset rot", command=self.__reset_rotation).place(relx=self.COMMON_X, rely=0.38, relheight=0.05, relwidth=0.095, anchor="ne")

	def __create_import_file_button(self):
		ttk.Button(self, text="Screenshot", command=self.__take_screenshot).place(relx=self.COMMON_X, rely=0.895, relheight=0.05, relwidth=0.1, anchor="ne")

	def __create_screenshot_button(self):
		self._file_name = tk.StringVar()
		ttk.Label(self, textvariable=self._file_name, foreground="#AAAAAA").place(relx=0.01, rely=0.96, relheight=0.035, relwidth=0.4)
		ttk.Button(self, text="Import file", command=self.__read_file).place(relx=self.COMMON_X, rely=0.815, relheight=0.05, relwidth=0.1, anchor="ne")

	def __create_up_down_left_right_buttons(self):
		# Common values for placements of the buttons
		COMM_X = 0.945
		COMM_Y = 0.6
		COMM_RELATIVE_HIEGHT = 0.04
		COMM_RELATIVE_WIDTH  = 0.03
		ttk.Button(self, text="U", command=self.__move_up).place(relx=COMM_X, rely=COMM_Y-0.04, relheight=COMM_RELATIVE_HIEGHT, relwidth=COMM_RELATIVE_WIDTH, anchor="ne")
		ttk.Button(self, text="D", command=self.__move_down).place(relx=COMM_X, rely=COMM_Y+0.04, relheight=COMM_RELATIVE_HIEGHT, relwidth=COMM_RELATIVE_WIDTH, anchor="ne")
		ttk.Button(self, text="L", command=self.__move_left).place(relx=COMM_X-0.035, rely=COMM_Y, relheight=COMM_RELATIVE_HIEGHT, relwidth=COMM_RELATIVE_WIDTH, anchor="ne")
		ttk.Button(self, text="R", command=self.__move_right).place(relx=COMM_X+0.035, rely=COMM_Y, relheight=COMM_RELATIVE_HIEGHT, relwidth=COMM_RELATIVE_WIDTH, anchor="ne")

	def __create_color_pickers(self):
		'''Create the color pickers to change line/canvas/fill colors'''
		COMM_Y = 0.935

		# FILL
		self.fill_color = tk.StringVar()
		self.fill_color.set("#000000")
		ttk.Label(self, text="Fill color:").place(relx=0.01, rely=COMM_Y-0.01, relheight=0.035)
		self._fill_btn = tk.Button(self, text="", command=self.__pick_color_fill, relief='flat')
		self._fill_btn.place(relx=0.08, rely=COMM_Y, relheight=0.015, relwidth=0.05)
		self._fill_btn['bg'] = self.fill_color.get()

		# LINE
		self.line_color = tk.StringVar()
		self.line_color.set("#0000FF")
		ttk.Label(self, text="Line color:").place(relx=0.2, rely=COMM_Y-0.01, relheight=0.035, anchor="ne")
		self._line_btn = tk.Button(self, text="", command=self.__pick_color_line, relief='flat')
		self._line_btn.place(relx=0.21, rely=COMM_Y, relheight=0.015, relwidth=0.05)
		self._line_btn['bg'] = self.line_color.get()

		# CANVAS' BACKGROUND
		ttk.Label(self, text="Canvas color:").place(relx=0.35, rely=COMM_Y - 0.01, relheight=0.035, anchor="ne")
		self._canvas_btn = tk.Button(self, text="", command=self.__pick_color_canvas, relief='flat')
		self._canvas_btn.place(relx=0.36, rely=COMM_Y, relheight=0.015, relwidth=0.05)
		self._canvas_btn['bg'] = self._canvas_color.get()

	def __create_fill_check(self):
		self._check_no_fill = tk.IntVar()
		ttk.Checkbutton(self, text="No fill", variable=self._check_no_fill, command=self.__changed, onvalue=True, offvalue=False).place(relx=0.80, rely=0.92)

	def __pick_color_fill(self):
		self.__pick_color("f")

	def __pick_color_line(self):
		self.__pick_color("l")

	def __pick_color_canvas(self):
		self.__pick_color("c")

	def __pick_color(self, picker):
		if(picker == "f"):
			col = colorchooser.askcolor(initialcolor = self.fill_color.get())
			if(col[1]):
				self.fill_color.set(col[1])
				self._fill_btn['bg']  = col[1]
		
		elif(picker == "c"):
			col = colorchooser.askcolor(initialcolor = self._canvas_color.get())
			if(col[1]):
				self._canvas_color.set(col[1])
				self._canvas_btn['bg']  = col[1]
				self._canvas['bg'] = self._canvas_color.get()
		
		else:
			col = colorchooser.askcolor(initialcolor = self.line_color.get())
			if(col[1]):
				self.line_color.set(col[1])
				self._line_btn['bg']  = col[1]
		
		self.__changed()

	def __mousewheel_scroll_in_canvas_up_event(self, *args):
		'''callback to the scrolling up in canvas event'''
		self._zoom_slider.set(self._zoom_slider.get()-0.5)

	def __mousewheel_scroll_in_canvas_down_event(self, *args):
		'''callback to the scrolling down in canvas event'''
		self._zoom_slider.set(self._zoom_slider.get()+0.5)

	def __continuous_x(self, *args):
		'''Callback to change event in checkbutton for continuous X rotation'''
		# Catch the disable-continuous-rotation-on-x event and update the rotation slider accoring to the last position
		if not self._check_x_continuos.get():
			x, _, _ = self._geometry_handler.orientation
			self.x_rotation_slider.set(x%math.pi)

	def __continuous_y(self, *args):
		'''Callback to change event in checkbutton for continuous Y rotation'''
		if not self._check_y_continuos.get():
			_, y, _ = self._geometry_handler.orientation
			self.y_rotation_slider.set(y%math.pi)

	def __continuous_z(self, *args):
		'''Callback to change event in checkbutton for continuous Z rotation'''
		if not self._check_z_continuos.get():
			_, _, z = self._geometry_handler.orientation
			self.z_rotation_slider.set(z%math.pi)

	def __get_canvas_shape(self):
		"""returns the shape of the canvas holding the visualized frame"""
		self.update()
		return self._canvas.winfo_width(), self._canvas.winfo_height()

	def __resized(self, *args):
		'''Callback to the window resize events'''
		w, h = self.__get_canvas_shape()
		if self._canvas_w != w or self._canvas_h != h:
			# Keep the object in the middle of the canvas
			self._geometry_handler.update_position((w-self._canvas_w)//2, (h-self._canvas_h)//2)
			self._canvas_w = w
			self._canvas_h = h
			self.__changed()

	def __changed(self, *args):
		'''Signal to the rendering function that something has changed in the object'''
		self._changed = True

	def __reset_rotation(self):
		self._geometry_handler.reset_rotation()
		self.x_rotation_slider.set(0)
		self.y_rotation_slider.set(0)
		self.z_rotation_slider.set(0)
		self.__changed()

	def __take_screenshot(self):
		#Grab the top-left corner's coordinates
		x = self.winfo_rootx() + self._canvas.winfo_x()
		y = self.winfo_rooty() + self._canvas.winfo_y()

		#Grab the bottom-right corner's coordinates
		x1 = x + self._canvas.winfo_width()
		y1 = y + self._canvas.winfo_height()

		#Crop the screenshot to the determined coordinates
		screenshot = ImageGrab.grab().crop((x, y, x1, y1))
		save_path = filedialog.asksaveasfilename(defaultextension=".png", filetypes=(("PNG Files", "*.png"), ("All Files", "*.*")))

		#Check if they actually saved and didn't exit before saving
		if save_path:
			screenshot.save(save_path)

	def __read_file(self):
		messagebox.showinfo(message='Only .obj files are compatible!', title="WARNING")

		file_path = filedialog.askopenfilename(defaultextension=".obj", filetypes=(("OBJ Files", "*.obj"), ("All Files", "*.*")))

		if len(file_path) and file_path[-4:] != ".obj":
			messagebox.showinfo(message="Incompatible file format", title="ERROR")

		elif len(file_path):
			self._file_name.set(file_path.split('/')[-1])
			self.__reset_rotation()
			with open(file_path) as file:
				self._geometry_handler.upload_object(*obj_files_handler.extract_data(file))
				self._file_exists = True

	def __move_up(self):
		self._geometry_handler.update_position(0, -1 * self.MOVING_STEP)
		self.__changed()

	def __move_down(self):
		self._geometry_handler.update_position(0, self.MOVING_STEP)
		self.__changed()

	def __move_left(self):
		self._geometry_handler.update_position(-1 * self.MOVING_STEP, 0)
		self.__changed()

	def __move_right(self):
		self._geometry_handler.update_position(self.MOVING_STEP, 0)
		self.__changed()

#	@time_me
	def render(self):
		'''Render the object on the screen'''
		# We need the continuous rotation to be enabled on one axis, and for the step to be non-zero
		continuous = ( (self.x_rotation_slider.get() and self._check_x_continuos.get()) 
		 			or (self.y_rotation_slider.get() and self._check_y_continuos.get()) 
		 			or (self.z_rotation_slider.get() and self._check_z_continuos.get())
		)

		if continuous: self.__step_rotations()
		self.__set_rotations()
		self.__set_zoom()

		if self._file_exists and (self._changed or continuous):
			#Delete all the previous points and lines in order to draw new ones
			self._canvas.delete("all")
			self.__update_colors()
			self.__draw_object()
			self._changed = False

	def __set_zoom(self):
		self._geometry_handler.set_zoom(self._zoom_slider.get())

	def __step_rotations(self):
		'''Increment the rotation angle by 1% on the enables axis'''
		self._geometry_handler.step_rotation(
			(self.x_rotation_slider.get()/100)*self._check_x_continuos.get(),
			(self.y_rotation_slider.get()/100)*self._check_y_continuos.get(),
			(self.z_rotation_slider.get()/100)*self._check_z_continuos.get(),
		)
	
	def __set_rotations(self):
		'''Set the required rotations for the geometry handler'''
		x, y, z = self._geometry_handler.orientation
		x_cont = self._check_x_continuos.get()
		y_cont = self._check_y_continuos.get()
		z_cont = self._check_z_continuos.get()
		self._geometry_handler.reset_rotation(
			x=self.x_rotation_slider.get()*(not x_cont) + x*x_cont, 
			y=self.y_rotation_slider.get()*(not y_cont) + y*y_cont, 
			z=self.z_rotation_slider.get()*(not z_cont) + z*z_cont
		)

	def __change_fill_color(self, color: str, no_fill: bool = False):
		'''Change the face fill color'''
		self._fill_color_holder = "" if no_fill else color

	def __change_line_color(self, color):
		''''''
		self._line_color_holder = color

	def __draw_point(self, point: 'tuple(int, int)') -> None:
		'''Draw a point on the canvas'''
		self._canvas.create_oval(point[0], point[1],
						   		 point[0], point[1],
						   		 width=self.POINT_SIZE,
						   		 fill=self.POINT_COLOR)

	@time_me
	def __draw_faces(self, points: dict) -> None:
		''''''
		for face in self._geometry_handler.faces:
			# Grab the points that make up that specific face
			to_draw = [points[f] for f in face]
			
			#for point in to_draw:
			#	if(point[0] < 0 or
			#	   point[1] < 0 or
			#	   point[0] > self.CANVAS_WIDTH or
			#	   point[1] > self.CANVAS_HEIGHT
			#	):
			#		continue # Don't draw points that are out of the screen
			#	# This is the slowest part of the GUI
			#	self.__draw_point(point)

			self._canvas.create_polygon(to_draw, outline=self._line_color_holder, fill=self._fill_color_holder)
	
	def __draw_object(self):
		'''Draw the object on the canvas'''
		projected_points = self._geometry_handler.transform_object()
		self.__draw_faces(projected_points)
	
	def __update_colors(self):
		self.__change_fill_color(self.fill_color.get(), self._check_no_fill.get())
		self.__change_line_color(self.line_color.get())



import pandas as pd
import mysql.connector
import os
import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime
import threading
import sys
import tkinter
import traceback
from math import radians, cos, sin, sqrt, atan2
import matplotlib.pyplot as plt
# import seaborn as sns
sys.path.append('../')

class Database():
    def __init__(self, selected_map = 'alamillo'):
        self.sensors=['Conductivity', 'PH', 'Sonar', 'Temperature', 'Turbidity']
        self.user="readonly"
        self.password="aceti"
        self.host="golem.us.es"
        self.selected_map = selected_map
        self.busy=False
        self.sensor_maps={}

    def query(self, date='2024-02-27', table="WQP", query = None, save=True,  database = "wqp"):
        if not self.busy:
            #execute in a thread
            # start=datetime.now()
            querythread=threading.Thread(target=self.__query,args=(date, table, query, save, database))
            self.busy=True
            querythread.start()
            # afterquery=datetime.now()
            # difference = afterquery - start
            # seconds_in_day = 24 * 60 * 60
            # aux=divmod(difference.days * seconds_in_day + difference.seconds, 60)
            # print(f"query delayed {aux[0]} minutes {aux[1]} seconds")

    def __query(self, date='2024-02-27', table="WQP", query=None, save =False, database = "wqp"): #Query can also be Positions
        self.date=date
        try:
            cnx = mysql.connector.connect(user=self.user, password=self.password, port="6006",
                                        host=self.host, database=database, connection_timeout=2)
        except:
            tkinter.messagebox.showerror(title="Mysql Connector", message="database connection failed, check your internet connection and whether US ports are open",)
            return
        # Crear un cursor
        cursor = cnx.cursor()
        # Definir la consulta SQL con el filtro de fecha
        # query = "SELECT * FROM ASV_US.ASV WHERE date(Date) >= '2024-02-16' AND date(Date) <= '2024-02-15'"
        if query is None:
            if table == "WQP":
                query = f"SELECT * FROM wqp.{table} where date(Date) = '{date}' AND Latitude<>0 AND Longitude<>0;"
            else:
                query = f"SELECT * FROM wqp.{table} where date(Date) = '{date}' AND Latitude<>0 AND Longitude<>0;"

        # Ejecutar la consulta
        try:
            cursor.execute(query)
        except:
            err=traceback.format_exc()
            print(f"query {query} failed: \n{err}")
            raise

        # Obtener nombres de columnas
        column_names = cursor.column_names

        # Obtener los resultados
        results = cursor.fetchall()

        # Cerrar el cursor y la conexión a la base de datos
        cursor.close()
        cnx.close()

        # Crear un DataFrame de pandas con los resultados
        self.df = pd.DataFrame(results, columns=column_names)

        if len(self.df)==0:
            self.busy=False
            print(f"empty dataframe at {date}")
            return

        print(f"queried {len(self.df)} rows")
        self.busy=False

    def resource_path(self, relative_path):
        """ Get absolute path to resource, works for dev and for PyInstaller """
        try:
            # PyInstaller creates a temp folder and stores path in _MEIPASS
            base_path = sys._MEIPASS
        except Exception:
            base_path = os.path.abspath(".")

        return os.path.join(base_path, relative_path)

    def refresh_dates(self, table="WQP", database="wqp"):
        try:
            cnx = mysql.connector.connect(user=self.user, password=self.password, port="6006",
                                        host=self.host, database=database, connection_timeout=2)
        except:
            tkinter.messagebox.showerror(title="Mysql Connector", message="database connection failed, check your internet connection and whether US ports are open",)
            return
        # Crear un cursor
        cursor = cnx.cursor()
        # Definir la consulta SQL con el filtro de fecha
        # query = "SELECT * FROM ASV_US.ASV WHERE date(Date) >= '2024-02-16' AND date(Date) <= '2024-02-15'"
        if table == "WQP":
            query = f"SELECT distinct DATE_FORMAT(Date, '%Y-%m-%d') FROM wqp.{table} WHERE Latitude<>0 AND Sensor='Conductivity' AND Data<>0;"
        elif table == "OBJECT_POSITION":
            query = f"SELECT distinct DATE_FORMAT(Date, '%Y-%m-%d') FROM wqp.{table} WHERE Latitude_Obj<>0;"
        elif table == "PATHS":
            query = f"SELECT distinct DATE_FORMAT(Date, '%Y-%m-%d') FROM wqp.{table} WHERE Latitude<>0;"
        else:
            query = f"SELECT distinct DATE_FORMAT(Date, '%Y-%m-%d') FROM {database}.{table};"

        # Ejecutar la consulta
        cursor.execute(query)

        # Obtener nombres de columnas
        column_names = cursor.column_names

        # Obtener los resultados
        results = cursor.fetchall()

        # Cerrar el cursor y la conexión a la base de datos
        cursor.close()
        cnx.close()

        # Crear un DataFrame de pandas con los resultados
        self.date_df = pd.DataFrame(results, columns=column_names)
        self.date_df = self.date_df[column_names[0]]
        if len(self.date_df)==0:
            self.busy=False
            print(f"there are no data from {table} at any date")
            return
        print(f"found {len(self.date_df)} dates with data")

    def refresh_maps(self, database="alec"):
        try:
            cnx = mysql.connector.connect(user=self.user, password=self.password, port="6006",
                                        host=self.host, database=database, connection_timeout=2)
        except:
            tkinter.messagebox.showerror(title="Mysql Connector", message="database connection failed, check your internet connection and whether US ports are open",)
            return
        # Crear un cursor
        cursor = cnx.cursor()
        query = f"SELECT distinct mapa FROM {database}.wqp_gaussian_maps;"

        # Ejecutar la consulta
        cursor.execute(query)

        # Obtener nombres de columnas
        column_names = cursor.column_names

        # Obtener los resultados
        results = cursor.fetchall()

        # Cerrar el cursor y la conexión a la base de datos
        cursor.close()
        cnx.close()

        # Crear un DataFrame de pandas con los resultados
        map_df = pd.DataFrame(results, columns=column_names)
        map_df = map_df["mapa"]
        if len(map_df)==0:
            self.busy=False
            print(f"there are no maps at any date")
            return
        print(f"found {len(map_df)} maps with data")
        return map_df.values
        
        


"""

Grid based Dijkstra planning

author: Atsushi Sakai(@Atsushi_twi)

"""

import math
import numpy as np
import time
# Dilation of the map scipy
from scipy.ndimage import binary_dilation

show_animation = False


class Dijkstra:

	def __init__(self, obstacle_map, robot_radius, obstacle = 1):
		"""
		Initialize map for a star planning

		ox: x position list of Obstacles [m]
		oy: y position list of Obstacles [m]
		resolution: grid resolution [m]
		rr: robot radius[m]
		"""

		self.min_x = None
		self.min_y = None
		self.max_x = None
		self.max_y = None
		self.x_width = None
		self.y_width = None
		self.obstacle_map = None

		self.robot_radius = robot_radius
		self.obstacle_map = obstacle_map.astype(bool)
		self.motion = self.get_motion_model()
		if obstacle == 1:
			self.obstacle_map = np.logical_not(self.obstacle_map)
		ox, oy = obstacle_map.nonzero()
		self.min_x = 0
		self.min_y = 0
		self.max_x = round(max(ox))
		self.max_y = round(max(oy))


		self.x_width = round((self.max_x - self.min_x))
		self.y_width = round((self.max_y - self.min_y))


	class Node:
		def __init__(self, x, y, cost, parent_index):
			self.x = x  # index of grid
			self.y = y  # index of grid
			self.cost = cost
			self.parent_index = parent_index  # index of previous Node

		def __str__(self):
			return str(self.x) + "," + str(self.y) + "," + str(
				self.cost) + "," + str(self.parent_index)

	def planning(self, start_node, goal_node, timeout = 10):
		"""
		dijkstra path search

		input:
			s_x: start x position [m]
			s_y: start y position [m]
			gx: goal x position [m]
			gx: goal x position [m]

		output:
			rx: x position list of the final path
			ry: y position list of the final path
		"""

		start_time = time.time()

		start_node = self.Node(self.calc_xy_index(start_node[0], self.min_x),
							   self.calc_xy_index(start_node[1], self.min_y), 0.0, -1)
		goal_node = self.Node(self.calc_xy_index(goal_node[0], self.min_x),
							  self.calc_xy_index(goal_node[1], self.min_y), 0.0, -1)

		open_set, closed_set = dict(), dict()
		open_set[self.calc_index(start_node)] = start_node

		while True:
			if not open_set:
				print("Cannot find path")
				return None
			c_id = min(open_set, key=lambda o: open_set[o].cost)
			current = open_set[c_id]


			if current.x == goal_node.x and current.y == goal_node.y:
				print("Find goal")
				goal_node.parent_index = current.parent_index
				goal_node.cost = current.cost
				break

			# Remove the item from the open set
			del open_set[c_id]

			# Add it to the closed set
			closed_set[c_id] = current

			# expand search grid based on motion model
			for move_x, move_y, move_cost in self.motion:
				node = self.Node(current.x + move_x,
								 current.y + move_y,
								 current.cost + move_cost, c_id)
				n_id = self.calc_index(node)

				if n_id in closed_set:
					continue

				if not self.verify_node(node):
					continue

				if n_id not in open_set:
					open_set[n_id] = node  # Discover a new node
				else:
					if open_set[n_id].cost >= node.cost:
						# This path is the best until now. record it!
						open_set[n_id] = node

			if time.time() - start_time > timeout:
				print("Timeout")
				return None

		rx,ry = self.calc_final_path(goal_node, closed_set)

		# Return as np array of two columns

		path = np.zeros((len(rx), 2))
		path[:, 0] = rx
		path[:, 1] = ry


		return path

	def calc_final_path(self, goal_node, closed_set):
		# generate final course
		rx, ry = [self.calc_position(goal_node.x, self.min_x)], [
			self.calc_position(goal_node.y, self.min_y)]
		parent_index = goal_node.parent_index
		while parent_index != -1:
			n = closed_set[parent_index]
			rx.append(self.calc_position(n.x, self.min_x))
			ry.append(self.calc_position(n.y, self.min_y))
			parent_index = n.parent_index

		return rx, ry

	def calc_position(self, index, minp):
		pos = index + minp
		return pos

	def calc_xy_index(self, position, minp):
		return round((position - minp))

	def calc_index(self, node):
		return (node.y - self.min_y) * self.x_width + (node.x - self.min_x)

	def verify_node(self, node):
		px = self.calc_position(node.x, self.min_x)
		py = self.calc_position(node.y, self.min_y)

		if px < self.min_x:
			return False
		if py < self.min_y:
			return False
		if px >= self.max_x:
			return False
		if py >= self.max_y:
			return False

		if self.obstacle_map[node.x][node.y]:
			return False

		return True

	def calc_obstacle_map(self, ox, oy):

		self.min_x = round(min(ox))
		self.min_y = round(min(oy))
		self.max_x = round(max(ox))
		self.max_y = round(max(oy))
		print("min_x:", self.min_x)
		print("min_y:", self.min_y)
		print("max_x:", self.max_x)
		print("max_y:", self.max_y)

		self.x_width = round((self.max_x - self.min_x))
		self.y_width = round((self.max_y - self.min_y))
		print("x_width:", self.x_width)
		print("y_width:", self.y_width)

		# obstacle map generation
		self.obstacle_map = [[False for _ in range(self.y_width)]
							 for _ in range(self.x_width)]


	@staticmethod
	def get_motion_model():
		# dx, dy, cost
		motion = [[1, 0, 1],
				  [0, 1, 1],
				  [-1, 0, 1],
				  [0, -1, 1],
				  [-1, -1, math.sqrt(2)],
				  [-1, 1, math.sqrt(2)],
				  [1, -1, math.sqrt(2)],
				  [1, 1, math.sqrt(2)]]

		return motion


def reduce_path_douglas_peucker(path, nav_map):

	""" Reduce the path using the Douglas-Peucker algorithm """

	from scipy.spatial import distance

	def perpendicular_distance(point, start, end):
		if np.array_equal(start, end):
			return distance.euclidean(point, start)
		else:
			num = abs((end[1] - start[1]) * point[0] - (end[0] - start[0]) * point[1] + end[0] * start[1] - end[1] * start[0])
			den = math.sqrt((end[1] - start[1])**2 + (end[0] - start[0])**2)
			return num / den

	def douglas_peucker(points, epsilon):
		if len(points) < 3:
			return points

		start = points[0]
		end = points[-1]

		max_dist = 0.0
		index = 0
		for i in range(1, len(points) - 1):
			dist = perpendicular_distance(points[i], start, end)
			if dist > max_dist:
				index = i
				max_dist = dist

		if max_dist > epsilon:
			left = douglas_peucker(points[:index + 1], epsilon)
			right = douglas_peucker(points[index:], epsilon)

			return np.vstack((left[:-1], right))
		else:
			return np.array([start, end])

	epsilon = 1.0  # Tolerance value, can be adjusted
	reduced_path = douglas_peucker(path, epsilon)

	return reduced_path

def reduce_path_bresenham(path, nav_map):

	""" Reduce the path by removing unnecessary waypoints if between two waypoints there is a straight line without obstacles """

	def is_line_free_bressen(p1, p2, nav_map):
		def bresenham(x0, y0, x1, y1):
			# yields integer coordinates on the line from (x0,y0) to (x1,y1)
			x0 = int(round(x0))
			y0 = int(round(y0))
			x1 = int(round(x1))
			y1 = int(round(y1))
			dx = abs(x1 - x0)
			dy = abs(y1 - y0)
			sx = 1 if x0 < x1 else -1
			sy = 1 if y0 < y1 else -1
			err = dx - dy
			x, y = x0, y0
			while True:
				yield x, y
				if x == x1 and y == y1:
					break
				e2 = 2 * err
				if e2 > -dy:
					err -= dy
					x += sx
				if e2 < dx:
					err += dx
					y += sy

		# ensure map is indexable as nav_map[row, col] -> nav_map[y, x]
		max_y, max_x = nav_map.shape
		x1, y1 = p1
		x2, y2 = p2

		for xi, yi in bresenham(x1, y1, x2, y2):
			# bounds check
			if xi < 0 or yi < 0 or xi >= max_x or yi >= max_y:
				# Out of bounds considered blocked
				return False
			# nav_map is expected to have 1 for obstacle
			if nav_map[yi, xi] == 1:
				return False

		return True
	
	# First and last points of the path
	index_i = 0
	index_f = len(path) - 1
	p1 = path[index_i]

	# New path will contain the reduced path from start to end
	new_path = [p1]
	N = len(path)
	if N <= 2:
		return np.asarray(path)

	# We'll use a greedy approach but find the farthest reachable index using binary search
	current_idx = 0

	while current_idx < N - 1:
		# binary search for farthest reachable index between (current_idx+1) .. (N-1)
		lo = current_idx + 1
		hi = N - 1
		best = lo
		while lo <= hi:
			mid = (lo + hi) // 2
			if is_line_free_bressen(path[current_idx], path[mid], nav_map):
				best = mid
				lo = mid + 1
			else:
				hi = mid - 1

		# If best didn't advance, force to next point to avoid infinite loops
		if best == current_idx:
			best = current_idx + 1

		new_path.append(path[best])
		current_idx = best

	return np.asarray(new_path)



	
def is_line_free(p1, p2, nav_map):
	""" Check if the line between two points is free of obstacles """

	# Get the coordinates of the line
	x1, y1 = p1
	x2, y2 = p2


	# Get the x and y coordinates of the line
	dx = x2 - x1
	dy = y2 - y1

	dt = 1000

	# Get the step size
	step_x = dx / dt
	step_y = dy / dt

	# Loop through the steps

	for t in range(dt):
		x = x1 + t * step_x
		y = y1 + t * step_y

		# Check if the point is inside an obstacle
		if nav_map[int(x), int(y)] == 1: # or nav_map[int(np.ceil(x)), int(np.ceil(y))] == 1:# or nav_map[int(np.ceil(x)), int(np.floor(y))] or nav_map[int(np.floor(x)), int(np.ceil(y))]: 
			return False
		
	return True

def is_line_free_new(p1, p2, nav_map):
    x1, y1 = map(int, map(round, p1))
    x2, y2 = map(int, map(round, p2))

    dx = abs(x2 - x1)
    dy = -abs(y2 - y1)
    sx = 1 if x1 < x2 else -1
    sy = 1 if y1 < y2 else -1
    err = dx + dy

    while True:
        # ojo al orden: nav_map[fila, columna] = nav_map[y, x]
        if not (0 <= y1 < nav_map.shape[0] and 0 <= x1 < nav_map.shape[1]):
            return False
        if nav_map[y1, x1]:   # celda ocupada
            return False
        if x1 == x2 and y1 == y2:
            break
        e2 = 2 * err
        if e2 >= dy:
            err += dy
            x1 += sx
        if e2 <= dx:
            err += dx
            y1 += sy
    return True


def reduce_path_new(path, nav_map):
    path = np.asarray(path)
    n = len(path)
    if n <= 2:
        return path

    new_path = [path[0]]
    i = 0
    while i < n - 1:
        progressed = False
        # probar del final hacia atrás
        for j in range(n - 1, i, -1):
            if is_line_free_new(path[i], path[j], nav_map):
                new_path.append(path[j])
                i = j
                progressed = True
                break
        if not progressed:
            # no hay línea directa: avanzar al siguiente para garantizar progreso
            new_path.append(path[i + 1])
            i += 1
    return np.asarray(new_path)

def reduce_path(path, nav_map):
	""" Reduce the path by removing unnecessary waypoints if between two waypoints there is a straight line without obstacles """

	# First and last points of the path
	index_i = 0
	index_f = len(path) - 1
	p1 = path[index_i]

	# New path
	new_path = [p1]
	path_completed = False
	# Loop through p1 to the furthest point in the path that is free of obstacles from p1

	while not path_completed:


		# Loop inverse from the end of the path to the next point of index index_i
		for i in reversed(range(index_i, len(path))):
			p2 = path[i] # Take the point

			# Check if the point is the next point of index index_i

			if index_i == i and len(new_path) > 1:
				p2 = path[i+1]
				new_path.append(p2)
				index_i = i
				p1 = p2
				break

			# Check if the line between p1 and p2 is free of obstacles
			elif is_line_free(p1, p2, nav_map):
				# If it is free, then we can remove p2
				new_path.append(p2)
				index_i = i
				p1 = p2

				# When p1 is the last point, then the path is completed
				if np.all(p1 == path[-1]):
					path_completed = True

				break

	
	# Return the new path reversed
	return np.asarray(new_path)[::-1]



def main():
	print(__file__ + " start!!")

	import numpy as np

	init_node = (7, 69)
	goal_node = (80, 209)
	#goal_node = (52, 73)
	# map_grid = np.loadtxt('/home/samuel/ASV_loyola_us_repo/ASV_Loyola_US/asv_workspace/mapas/othermap.txt')
	

	map_grid = np.loadtxt('mapas/Alamillo95x216plantilla.csv', delimiter=" ").astype(int)

	# Dilation of the map
	dilated_map = binary_dilation(map_grid, np.ones((3, 3)))
	
	# set obstacle positions
	ox, oy = np.where(map_grid == 1)

	dijkstra = Dijkstra(dilated_map, 1)
	path = dijkstra.planning(init_node, goal_node)

	if path is not None:
		refined_path = reduce_path(path, dilated_map)
		print(refined_path)


if __name__ == '__main__':
	main()